/*
	command support

	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2014

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef _INCLUDE_COMMAND_H_
#define _INCLUDE_COMMAND_H_

#include <time.h>
#include <stdbool.h>
#include <sys/select.h>
#include <stdint.h>


enum command_ptype {
	COMMAND_PTYPE_NULL = 0,
	COMMAND_PTYPE_LIST = 1,
	COMMAND_PTYPE_NAME = 2,
	COMMAND_PTYPE_ID = 3,
	COMMAND_PTYPE_SETPOINT = 4, /* Simple direct setpoint */
	COMMAND_PTYPE_SPEED = 5, /* Simple direct 1st derivative setpoint */
	COMMAND_PTYPE_SETPOINT_TIME = 6, /* Setpoint to be reached @ time */
	
	COMMAND_PTYPE_MAX, /* Always the last entry! */
};

char *enum_command_ptype2str(enum command_ptype ptype);

enum command_value_type {
	COMMAND_VALUE_TYPE_FLOAT,	/* 4 bytes */
	COMMAND_VALUE_TYPE_BOOL,	/* 1 byte */
	COMMAND_VALUE_TYPE_UINT8,
	COMMAND_VALUE_TYPE_UINT16,
	COMMAND_VALUE_TYPE_UINT32,
	COMMAND_VALUE_TYPE_SINT8,
	COMMAND_VALUE_TYPE_SINT16,
	COMMAND_VALUE_TYPE_SINT32,
};

char *enum_command_value_type2str(enum command_value_type type);

enum command_state {
	COMMAND_STATE_DISCONNECTED,
	COMMAND_STATE_CONNECTED,
	COMMAND_STATE_READY,
	COMMAND_STATE_RECEIVING,
};

char *enum_command_state2str(enum command_state state);

struct command_entry {
	enum command_ptype type;
	union {
		float f;
		bool b;
		uint8_t u8;
		uint16_t u16;
		uint32_t u32;
		int8_t s8;
		int16_t s16;
		int32_t s32;
	} value;
	struct timespec t;
};

struct command {
	int fd;
	char *host;
	int port;
	
	unsigned char *rx_buffer;
	size_t rx_len;
	size_t rx_buf_size;
	
	size_t packet_size;
	
	char *name;
	char *unit;
	char *id;
	enum command_value_type type;

	bool list_received;
	bool recover;
	time_t recovertime;

	void (*handler_list_entry)(struct command *, 
	    char *name, enum command_value_type type, char *unit,
	    int typec, enum command_ptype typev[]);
	void (*handler_name)(struct command *, char *name);
	void (*handler_id)(struct command *, char *id);
	void (*handler_entry)(struct command *, struct command_entry *entry);
	void (*handler_close)(struct command *);
	
	void *private;
};

struct command_pkt {
	enum command_ptype type;
	unsigned char *data;
	size_t len;
	size_t buflen;
};


struct command *command_alloc(void);
void command_free(struct command *command);
struct command *command_open(char *host, int port);
void command_initialize(struct command *command, size_t buffer);
void command_initialize_fd(struct command *command, int fd);
void command_close(struct command *command);

struct command *command_open_simple(char *host, int port, char *name, char *id,
    enum command_value_type type);
int command_send(struct command *command, struct command_entry *entry);

int command_name_set(struct command *command, char *name);
int command_id_set(struct command *command, char *id);

bool command_fd_set(struct command *command, fd_set *set, int *high);
int command_handle(struct command *command, fd_set *set);
enum command_state command_state_get(struct command *command);
void command_autorecover(struct command *command, bool value);

bool command_fd_read(struct command * command);

struct command_pkt *command_packet_get(struct command *command);
void command_packet_put(struct command_pkt *pkt);

void command_packet_initialize(struct command_pkt *pkt);
int command_packet_write(struct command *command, struct command_pkt *pkt);

struct command_pkt *command_packet_new(void);
int command_packet_resize(struct command_pkt *pkt, size_t len);
int command_packet_list_add(struct command_pkt *list_pkt, 
    char *name, enum command_value_type type, char *unit,
    int typec, enum command_ptype typev[]);
    
#endif /* _INCLUDE_COMMAND_H_ */
