/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include <trace/trace.h>

time_t starttime;

#define VAL_BUFLEN (60 * 5)
FILE *fdplot = NULL;
int doplot = 0;

struct trace_view {
	struct trace *trace;
	char *tmpfilename;
	FILE *fdtmp;
	struct trace_value value[VAL_BUFLEN];
	int bufpos;
};

struct trace_view *traces;


void plot_values(struct trace_view *traces, int nr_traces)
{
	int i;
	int bufpos = (traces[0].bufpos+VAL_BUFLEN-1) % VAL_BUFLEN;
	long long now = traces[0].value[bufpos].t.tv_sec;
	long nown = traces[0].value[bufpos].t.tv_nsec;
	
	fprintf(fdplot, "plot ");

	for(i = 0; i < nr_traces; i++) {
		if (i > 0)
			fprintf(fdplot, ", ");
		fprintf(fdplot, "'%s' using ($1-%lld.%09ld):($2) title '%s [%s]' with lines",
		    traces[i].tmpfilename,
		    now, nown,
		    traces[i].trace->name, 
		    "");
	}
	fprintf(fdplot, "\n");
	fflush(NULL);

	return;
}

static void print_value(FILE *fd,
    struct trace *trace, struct trace_value *value)
{
	switch (trace->type) {
		case TRACE_VALUE_TYPE_FLOAT:
			fprintf(fd, "%ld.%09ld %e\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.f);
			break;
		case TRACE_VALUE_TYPE_BOOL:
			fprintf(fd, "%ld.%09ld %d\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.b);
			break;
		case TRACE_VALUE_TYPE_UINT8:
			fprintf(fd, "%ld.%09ld %u\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.u8);
			break;
		case TRACE_VALUE_TYPE_UINT16:
			fprintf(fd, "%ld.%09ld %u\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.u16);
			break;
		case TRACE_VALUE_TYPE_UINT32:
			fprintf(fd, "%ld.%09ld %u\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.u32);
			break;
		case TRACE_VALUE_TYPE_SINT8:
			fprintf(fd, "%ld.%09ld %d\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.s8);
			break;
		case TRACE_VALUE_TYPE_SINT16:
			fprintf(fd, "%ld.%09ld %d\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.s16);
			break;
		case TRACE_VALUE_TYPE_SINT32:
			fprintf(fd, "%ld.%09ld %d\n", 
			value->t.tv_sec, value->t.tv_nsec, value->value.s32);
			break;
	}
}

static void handler_value(struct trace *trace,
    struct trace_value *value)
{
	struct trace_view *view;
	int i;
	
	view = trace->private;
	
	view->value[view->bufpos] = *value;

	freopen(NULL, "w", view->fdtmp);
	for (i = view->bufpos + 1; i < VAL_BUFLEN; i++) {
		if (!view->value[i].t.tv_sec)
			continue;

		print_value(view->fdtmp, trace, &view->value[i]);
	}
	for (i = 0; i <= view->bufpos; i++) {
		if (!view->value[i].t.tv_sec)
			continue;

		print_value(view->fdtmp, trace, &view->value[i]);
	}
	fflush(view->fdtmp);


	view->bufpos++;
	view->bufpos %= VAL_BUFLEN;
	
	doplot++;
}

int main(int argc, char **argv)
{
        int i, ret = 0;

	starttime = time(NULL);

	traces = calloc(sizeof(struct trace_view), argc - 1);
	for (i = 0; i < argc - 1; i++) {
		char *tracename = argv[i+1];
		struct trace_pkt *pkt;
		struct timespec t_int;
		int fdmkstemp;
		
		printf("%s\n", tracename);
		traces[i].trace = trace_open("localhost", 10000);
		if (!traces[i].trace) {
			printf("open trace failed\n");
			return 1;
		}
		traces[i].trace->private = &traces[i];
		trace_name_set(traces[i].trace, tracename);

		traces[i].tmpfilename = malloc(100);
		sprintf(traces[i].tmpfilename, "/tmp/trace_view_XXXXXX");
		fdmkstemp = mkstemp(traces[i].tmpfilename);
	
		printf("Going to use %s temp file for '%s' data\n",
		    traces[i].tmpfilename, tracename);
		traces[i].fdtmp = fdopen(fdmkstemp, "w");
		if (traces[i].fdtmp == NULL) {
			perror("fopen() failed");
			return 1;
		}

		t_int.tv_sec = 0;
		t_int.tv_nsec = 200*1000*1000;

		pkt = trace_packet_new();
		trace_packet_interval_set(pkt, 
		    &t_int, TRACE_INTERVAL_TYPE_INTERVAL);
		trace_packet_write(traces[i].trace, pkt);
		trace_packet_put(pkt);

		traces[i].trace->handler_value = handler_value;
	}
	
	fdplot = popen("gnuplot -noraise", "w");
	if (fdplot == NULL) {
		perror("popen() failed");
		return 1;
	}
	
	fprintf(fdplot, "set grid\n");

	do {
		fd_set fdrx;
		
		int high = 0;
		
		FD_ZERO(&fdrx);
	
		for (i = 0; i < argc - 1; i++)
			trace_fd_set(traces[i].trace, &fdrx, &high);
		
		select(high+1, &fdrx, NULL, NULL, NULL);

		for (i = 0; i < argc - 1; i++)
			trace_handle(traces[i].trace, &fdrx);

		if (doplot >= argc - 1) {
			plot_values(traces, argc - 1);
			doplot = 0;
		}
	} while (ret == 0);

	for (i = 0; i < argc - 1; i++)
		unlink(traces[i].tmpfilename);

	return 0;
}
