/*
	Command generator for manual speed control

	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2008, 2014
	Copyright Stichting C.A. Muller Radioastronomiestation, 2008

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <ctype.h>
#include <inttypes.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <command/command.h>

#include <utils/command_server.h>

char *command_host = "localhost";
int command_port = 11000;
int cmd_port = 11051;
char *az_command_spg = "Azimuth_Setpoint";
char *el_command_spg = "Elevation_Setpoint";

int go = 0;
float sp_az = 0;
float sp_el = 0;


int handle_cmd(char *command)
{
	char *aaz, *ael, *ptr;
	
	aaz = strtok_r(command, " \t", &ptr);
	ael = strtok_r(NULL, " \t", &ptr);
	
	if (!aaz || !ael)
		return -1;
	
	while (ael[strlen(ael)-1] == '\n' ||
	    ael[strlen(ael)-1] == '\r')
		ael[strlen(ael)-1] = 0;
	
	sscanf(aaz, "%f", &sp_az);
	sscanf(ael, "%f", &sp_el);
	
	go = 1;
	
	return 0;
}


int main(int argc, char **argv)
{
	struct command *sp_command_az = NULL;
	struct command *sp_command_el = NULL;
	struct command_server *cmd_srv;

	time_t lastt = 0;

	signal(SIGPIPE, SIG_IGN);

	if (argc == 2 || argc == 4) {
		command_host = argv[argc - 1];
	}

	sp_command_az = command_open_simple(command_host, command_port,
	    az_command_spg, "console/manual", COMMAND_VALUE_TYPE_FLOAT);
	sp_command_el = command_open_simple(command_host, command_port,
	    el_command_spg, "console/manual", COMMAND_VALUE_TYPE_FLOAT);
	if (!sp_command_az || !sp_command_el) {
		fprintf(stderr, "Setpoint generator(s) not found\n");
		exit(-1);
	}

	cmd_srv = command_server_create(cmd_port, 0, 100);
	if (!cmd_srv) {
		printf("Could not open listen port for commands\n");
	} else {
		command_server_handler_raw_set(cmd_srv, handle_cmd);
	}

	while (1)
	{
		time_t t;
		int high = 0;
		fd_set fdset_rx;
		struct timeval tv;
		
		t = time(NULL);
		if (t != lastt) {
			lastt = t;
		
			if (go) {
				struct command_entry entry;
				printf("Speed: %e %e\n", sp_az, sp_el);

				entry.type = COMMAND_PTYPE_SPEED;
				entry.value.f = sp_az;
				command_send(sp_command_az, &entry);
				command_send(sp_command_az, &entry);
				entry.value.f = sp_el;
				command_send(sp_command_el, &entry);
				command_send(sp_command_el, &entry);
				go = 0;
			}
		}

		FD_ZERO(&fdset_rx);
	
		command_server_fdset_add(cmd_srv, &fdset_rx, &high);
		
		tv.tv_sec = 0;
		tv.tv_usec = 10000;

		select(high, &fdset_rx, NULL, NULL, &tv);

		command_server_fdset_handle(cmd_srv, &fdset_rx);
	}

}
