/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>

#include <trace/trace.h>
#include <websocket_plugin.h>

#include <utils/tcp_connect.h>

#include <dt_port_numbers.h>
#include <utils/dt_host.h>

struct plugin_private {
	const WebSocketServer *server;

	pthread_t *threads;
	int nr_threads;
	
	bool connected;
};

static void *on_connect(const WebSocketServer *server)
{
	struct plugin_private *priv;
	
	priv = calloc(sizeof(struct plugin_private), 1);

	priv->server = server;
	priv->connected = true;
	
	return priv;
}

static void handler_trace_value(struct trace *trace,
    struct trace_value *value)
{
	struct plugin_private *priv;
	char *msg;

	priv = trace->private;

	switch (trace->type) {
		case TRACE_VALUE_TYPE_FLOAT:
			asprintf(&msg, "trace %s %ld.%09ld %e\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.f);
			break;
		case TRACE_VALUE_TYPE_BOOL:
			asprintf(&msg, "trace %s %ld.%09ld %d\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.b);
			break;
		case TRACE_VALUE_TYPE_UINT8:
			asprintf(&msg, "trace %s %ld.%09ld %u\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.u8);
			break;
		case TRACE_VALUE_TYPE_UINT16:
			asprintf(&msg, "trace %s %ld.%09ld %u\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.u16);
			break;
		case TRACE_VALUE_TYPE_UINT32:
			asprintf(&msg, "trace %s %ld.%09ld %u\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.u32);
			break;
		case TRACE_VALUE_TYPE_SINT8:
			asprintf(&msg, "trace %s %ld.%09ld %d\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.s8);
			break;
		case TRACE_VALUE_TYPE_SINT16:
			asprintf(&msg, "trace %s %ld.%09ld %d\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.s16);
			break;
		case TRACE_VALUE_TYPE_SINT32:
			asprintf(&msg, "trace %s %ld.%09ld %d\n", 
			trace->name,
			value->t.tv_sec, value->t.tv_nsec, value->value.s32);
			break;
	}
	priv->server->send(priv->server, MESSAGE_TYPE_TEXT, 
	    (unsigned char *)msg, strlen(msg));
	free(msg);
}

void *trace_thread(void *arg)
{
	struct trace *trace = arg;
	struct plugin_private *priv = trace->private;
	
	trace->handler_value = handler_trace_value;

	while (priv->connected) {
		fd_set fdrx;
		int high;
		struct timeval to;
		
		FD_ZERO(&fdrx);
		
		if (!trace_fd_set(trace, &fdrx, &high)) {
			sleep(1);
			continue;
		}
		
		to.tv_sec = 1;
		to.tv_usec = 0;
		select(high + 1, &fdrx, NULL, NULL, &to);
		
		trace_handle(trace, &fdrx);
	}
	
	trace_free(trace);

	return NULL;
}

static void start_trace(struct plugin_private *priv, int freq, char *variable)
{
	struct trace *trace;
	struct timespec t_int;
	enum trace_interval_type type;

	trace = trace_open("localhost", 10000);
	

	trace_name_set(trace, variable);
	free(variable);
	
	trace_autorecover(trace, true);
	
	trace->private = priv;
	
	if (freq) {
		uint64_t nsec;
		
		nsec = 1000000000 / freq;
		
		type = TRACE_INTERVAL_TYPE_INTERVAL;
		t_int.tv_sec = nsec / 1000000000;
		t_int.tv_nsec = nsec - t_int.tv_sec * 1000000000;
		t_int.tv_sec = 0;
		t_int.tv_nsec = 500000000;
		
	} else {
		type = TRACE_INTERVAL_TYPE_CHANGED;
		t_int.tv_sec = 0;
		t_int.tv_nsec = 0;
	}
	
	trace_interval_set(trace, &t_int, type);

	priv->threads = realloc(priv->threads, 
	     sizeof(pthread_t) * (priv->nr_threads + 1));
	
	pthread_create(&priv->threads[priv->nr_threads], NULL, 
	    trace_thread, trace);
	
	priv->nr_threads++;
}

struct status {
	char *ident;
	struct plugin_private *priv;
};

#define LINEBUF_SIZE 8192
void *status_thread(void *arg)
{
	struct status *status = arg;
	char *ident = status->ident;
	struct plugin_private *priv = status->priv;
	char *host;
	int port = -1;
	int fd = -1;
	int pos = 0;
	unsigned char linebuf[LINEBUF_SIZE];
	
	free(status);
	
	host = dt_host_console();
	
	if (!strcmp(ident, "sat"))
		port = CONSOLE_SAT_STAT_PORT;
	else if (!strcmp(ident, "sun"))
		port = CONSOLE_SUN_STAT_PORT;
	else if (!strcmp(ident, "moon"))
		port = CONSOLE_MOON_STAT_PORT;
	else if (!strcmp(ident, "auth"))
		port = CONSOLE_STATUS_PORT;
	else if (!strcmp(ident, "j2000"))
		port = CONSOLE_J2000_POS_PORT;
	else if (!strcmp(ident, "log"))
		port = CONSOLE_LOG_PORT;
	else if (!strcmp(ident, "weather"))
		port = CONSOLE_WEATHER_PORT;
	else if (!strcmp(ident, "dt_model"))
		port = CONSOLE_DT_MODEL_STAT_PORT;



	while (priv->connected) {
			char *msg;
			int r;
			fd_set fd_rd;
			struct timeval timeout;

			if (fd < 0) {
				fd = tcp_connect(host, port);
				if (fd < 0) {
					sleep(1);
					continue;
				}	
			}

			timeout.tv_usec = 0;
			timeout.tv_sec = 10;

			FD_ZERO(&fd_rd);
			FD_SET(fd, &fd_rd);
			select(fd + 1, &fd_rd, NULL, NULL, &timeout);	
		
			if (FD_ISSET(fd, &fd_rd)) {
				r = read(fd, linebuf + pos, LINEBUF_SIZE - 1 - pos);
				if (r > 0) {
					int i;
					
					for (i = 0; i <r; i++) {
						if (linebuf[pos + i] == '\n') {
							linebuf[pos + i] = 0;
							asprintf(&msg, 
							    "status %s %s\n",
							    ident, linebuf);
							priv->server->send(
							    priv->server, 
							    MESSAGE_TYPE_TEXT,
							    (unsigned char *)msg, 
							    strlen(msg));
							free(msg);
							if (r - i - 1 > 0)
								memmove(
								    linebuf, 
								    linebuf + pos + i + 1, 
								    r - i - 1);
							pos -= pos + i +1;
						}
					}
					pos += r;
				} else if (errno != EAGAIN) {
					close(fd);
					fd = -1;
					sleep(1);
				}
			}
	}
	free(ident);
	
	return NULL;
}

static void start_status(struct plugin_private *priv, char *ident)
{
	struct status *status;
	
	status = malloc(sizeof(struct status));
	status->priv = priv;
	status->ident = ident;

	priv->threads = realloc(priv->threads, 
	     sizeof(pthread_t) * (priv->nr_threads + 1));

	pthread_create(&priv->threads[priv->nr_threads], NULL, 
	    status_thread, status);
	
	priv->nr_threads++;	
}

static void do_command(char *ident, char *command)
{
	char *host;
	int port = -1;
	int fd;
	
	host = dt_host_console();
	
	if (!strcmp(ident, "sat"))
		port = CONSOLE_SAT_CMD_PORT;
	else if (!strcmp(ident, "sun"))
		port = CONSOLE_SUN_CMD_PORT;
	else if (!strcmp(ident, "moon"))
		port = CONSOLE_MOON_CMD_PORT;
	else if (!strcmp(ident, "auth"))
		port = CONSOLE_ACTIVE_PORT;
	else if (!strcmp(ident, "j2000"))
		port = CONSOLE_J2000_CMD_PORT;
	else if (!strcmp(ident, "offset"))
		port = CONSOLE_OFFSET_PORT;
	else if (!strcmp(ident, "azel"))
		port = CONSOLE_AZEL_PORT;
	else if (!strcmp(ident, "manual"))
		port = CONSOLE_MANUAL_CMD_PORT;
	else if (!strcmp(ident, "dt_model"))
		port = CONSOLE_DT_MODEL_CMD_PORT;

	fd = tcp_connect(host, port);
	
	if (fd >= 0) {
		ioctl(fd, FIONBIO, &(int){ 0 });
		
		write(fd, command, strlen(command));

		write(fd,"\n", 1);
		fsync(fd);
		tcp_flushwrite(fd);

		close(fd);
	}
}

static void do_shell(char *command)
{
	char *host;
	int port = -1;
	int fd;
	
	host = dt_host_controller();
	
	port = CTRL_SHELL_PORT;

	fd = tcp_connect(host, port);
	
	if (fd >= 0) {
		ioctl(fd, FIONBIO, &(int){ 0 });
		
		write(fd, command, strlen(command));
 		fsync(fd);
		
		write(fd,"\n", 1);
		write(fd,"quit\n", 5);
 		fsync(fd);
		tcp_flushwrite(fd);
		tcp_flushread(fd);

		close(fd);
	}
}

static size_t on_message(void *plugin_private,
                    const WebSocketServer *server,
                    const int type,
                    unsigned char *buffer,
                    const size_t buffer_size)
{
	struct plugin_private *priv = plugin_private;
	size_t ret = 0;
	char text[buffer_size + 1];
	char *cmd;
	
	if (type != MESSAGE_TYPE_TEXT)
		return 0;
	
	memcpy(text, buffer, buffer_size);
	text[buffer_size] = 0;

	sscanf(text, "%as ", &cmd);
	
	if (!strcmp(cmd, "trace")) {
		int freq;
		char *variable;
		
		sscanf(text, "%*s %d %as", &freq, &variable);

		start_trace(priv, freq, variable);
	} else if (!strcmp(cmd, "command")) {
		char *ident;
		char *command;
		
		sscanf(text, "%*s %as", &ident);
		command = text + strlen(cmd) + strlen(ident) + 2;
		do_command(ident, command);

		free(ident);
	} else if (!strcmp(cmd, "shell")) {
		char *command;
		
		command = text + strlen(cmd) + 1;
		do_shell(command);
	} else if (!strcmp(cmd, "status")) {
		char *ident;
		
		sscanf(text, "%*s %as", &ident);
		start_status(priv, ident);
	} else {
		server->send(priv->server, MESSAGE_TYPE_TEXT, 
	    		buffer, buffer_size);		
	}
		
	free(cmd);
	return ret;
}

static void on_disconnect(void *plugin_private, const WebSocketServer *server)
{
	struct plugin_private *priv = plugin_private;
	int i;
	
	priv->connected = false;
	
	for (i = 0; i < priv->nr_threads; i++) {
		pthread_join(priv->threads[i], NULL);
	}
	
	free(priv);
}


static WebSocketPlugin s_plugin = {
	size: sizeof(WebSocketPlugin),
	version: WEBSOCKET_PLUGIN_VERSION_0,
	destroy: NULL,
	on_connect: on_connect,
	on_message: on_message,
	on_disconnect: on_disconnect,
};

extern EXPORT WebSocketPlugin * CALLBACK dt_init()
{
  return &s_plugin;
}
