/*
	Command authoriser

	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2008, 2009, 2013, 2014
	Copyright Stichting C.A. Muller Radioastronomiestation, 2008, 2009, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <inttypes.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <command/command.h>
#include <utils/tcp_listen.h>
#include <utils/status_server.h>
#include <utils/command_server.h>

#include <dt_port_numbers.h>
#include <utils/dt_host.h>

#include <log/log.h>


struct spg_client {
	struct spg_client *next;
	int cleanup;
	int pending;
	struct command *command;
	struct command_entry entry;
	struct spg *spg;
};

struct spg {
	struct command *spg_cmd;
	char *name;
};

struct spg *spgs = NULL;
static int spgs_nr = 0;

struct spg_client *clients = NULL;

typedef union {
	float f;
	uint32_t u;
} float32_t;



char *command_host = "localhost";
int future_hold = 3;
char *command_id = "console/spg_auth";

char *logfile = "spg_auth.log";

float az_offset = 0.0;
float el_offset = 0.0;

char current_auth[100] = { "none" };


int fd_listen;
int fd_console;
int fd_offset;

static struct command_pkt *list_pkt;


static void handler_name(struct command *command, char *name)
{
	struct spg_client *client = command->private;
	int i;
	
	for (i = 0; i < spgs_nr; i++) {
		if (!strcmp(spgs[i].name, name)) {
			client->spg = &spgs[i];
			break;
		}
	}
}


int handle_offsets(char *command)
{
	sscanf(command, "%f %f", &az_offset, &el_offset);
	log_send(LOG_T_DEBUG, "New offset: %e %e\n", az_offset, el_offset);

	return 0;
}

void try_command(struct command *command)
{
	struct spg_client *client = command->private;
	struct command_entry *entry = &client->entry;
	
	if (!client->command->id || !client->spg || strcmp(client->command->id, current_auth)) {
		client->pending = 0;
		return;
	}

	if (entry->type == COMMAND_PTYPE_SETPOINT_TIME) {
		/* Check if setpoint-time command is not to far in the future */
		if  (entry->t.tv_sec > time(NULL) + future_hold)
			return;
		
	}

	if (!strcmp(client->spg->name, "Azimuth_Setpoint")) {
		switch(entry->type) {
			case COMMAND_PTYPE_SETPOINT:
			case COMMAND_PTYPE_SETPOINT_TIME:
				entry->value.f += az_offset;
			default:
				break;
		}
	}
	if (!strcmp(client->spg->name, "Elevation_Setpoint")) {
		switch(entry->type) {
			case COMMAND_PTYPE_SETPOINT:
			case COMMAND_PTYPE_SETPOINT_TIME:
				entry->value.f += el_offset;
			default:
				break;
		}
	}
	command_send(client->spg->spg_cmd, entry);
//	log_command(client->spg, client->cmd, client->cmdlen);

	client->pending = 0;
	log_send(LOG_T_DEBUG, "Done command");
}

void handler_entry(struct command *command, struct command_entry *entry)
{
	struct spg_client *client = command->private;
	
	memcpy(&client->entry, entry, sizeof(struct command_entry));
	client->pending = 1;
	
	try_command(command);
}

void handler_close(struct command *command)
{
	struct spg_client *client = command->private;
	
	log_send(LOG_T_DEBUG, "Client connection lost");
	
	client->cleanup = true;
}

void new_connection(void)
{
	int fd;
	struct spg_client *new_client;
	
	fd = tcp_accept(fd_listen);
	if (fd < 0)
		return;
	ioctl(fd, FIONBIO, &(int){ 1 });
	
	new_client = calloc(1, sizeof(struct spg_client));
	if (!new_client)
		goto err_client;
	
	new_client->command = command_alloc();
	if (!new_client->command)
		goto err_command;
		
	command_initialize_fd(new_client->command, fd);
	
	command_packet_write(new_client->command, list_pkt);
	new_client->command->handler_name = handler_name;
	new_client->command->handler_entry = handler_entry;
	new_client->command->handler_close = handler_close;
	new_client->command->private = new_client;
	
	new_client->next = clients;
	clients = new_client;

	log_send(LOG_T_DEBUG, "New client");
	return;

err_command:
	free(new_client);
err_client:
	close(fd);
	
}



int handle_con_client(char *command)
{
	if (strcmp(current_auth, command)) {
		int i;
						
		strcpy(current_auth, command);
		while (current_auth[strlen(current_auth)-1] == '\n' ||
		    current_auth[strlen(current_auth)-1] == '\r')
			current_auth[strlen(current_auth)-1] = 0;

		for (i = 0; i < spgs_nr; i++) {
			struct command_entry entry;
			
			entry.type = COMMAND_PTYPE_SPEED;
			entry.value.f = 0.0;
			command_send(spgs[i].spg_cmd, &entry);
		}

		log_send(LOG_T_INFO, "Changed setpoint source to: %s", 
		    current_auth);
	}
	return 0;
}

void update_status(struct status_server *stat_srv)
{
	static time_t last = 0;
	time_t now;
	char status[100];
	
	now = time(NULL);
	if (now == last)
		return;
	
	last = now;
	
	snprintf(status, 99, "%s %e %e\n", current_auth, az_offset, el_offset);
	status[99] = 0;
	
	status_server_send(stat_srv, status);
}

static void handler_list_entry(struct command *command, 
    char *name, enum command_value_type type, char *unit,
    int typec, enum command_ptype typev[])
{
	command_packet_list_add(list_pkt, name, type, unit, typec, typev);

	spgs_nr++;
	spgs = realloc(spgs, spgs_nr * sizeof(struct spg));
	spgs[spgs_nr - 1].name = strdup(name);
	
	command_autorecover(command, false);
}

int main (int argc, char **argv)
{
	int i;
	struct status_server *stat_srv;
	struct command_server *offset_srv;
	struct command_server *console_srv;
	struct command *list;

	signal(SIGPIPE, SIG_IGN);

	command_host = dt_host_controller();

	if (argc > 1) {
		logfile = argv[2];
	}

	log_client_start(dt_host_console(), CONSOLE_LOG_PORT_IN, 
	    LOG_T_DEBUG, LOG_T_INFO, "spg_auth");

	log_send(LOG_T_INFO, "spg_auth started");

	/* Get list of all traces */
	log_send(LOG_T_DEBUG, "Going to retrieve list of traces");

	list = command_open(command_host, CTRL_COMMAND_PORT);
	list->handler_list_entry = handler_list_entry;
	command_autorecover(list, true);

	list_pkt = command_packet_new();

	while (command_state_get(list) != COMMAND_STATE_READY) {
		fd_set fd_rd;
		struct timeval timeout;
		int high = 0;

		timeout.tv_sec = 1;
		timeout.tv_usec = 10;
		
		FD_ZERO(&fd_rd);
		command_fd_set(list, &fd_rd, &high);	
		select(high + 1, &fd_rd, NULL, NULL, &timeout);
		command_handle(list, &fd_rd);
	
		/* Prevent cpu hogging */
		if (command_state_get(list) != COMMAND_STATE_READY)
			sleep(1);
//		printf("state: %d\n", command_state_get(list));
	}
	command_close(list);
	command_free(list);

	log_send(LOG_T_DEBUG, "Recieved list from server, now waiting for clients");
	for (i = 0; i < spgs_nr; i++) {
		spgs[i].spg_cmd = command_open_simple(
		    command_host, CTRL_COMMAND_PORT, spgs[i].name, command_id,
		    COMMAND_VALUE_TYPE_FLOAT);
	}

	fd_listen = tcp_listen(CONSOLE_COMMAND_PORT, 0, 100);
	if (fd_listen < 0) {
		log_send(LOG_T_ERROR, "Could not open listen port for commands");
		fd_listen = 0;
	}
	ioctl(fd_listen, FIONBIO, &(int){ 1 });

	console_srv = command_server_create(CONSOLE_ACTIVE_PORT, 0, 100);
	if (!console_srv) {
		log_send(LOG_T_ERROR, "Could not open listen port for consoles");
	}
	stat_srv = status_server_create(CONSOLE_STATUS_PORT, 0, 100);
	if (!stat_srv) {
		log_send(LOG_T_ERROR, "Could not open listen port for status");
	}
	
	offset_srv = command_server_create(CONSOLE_OFFSET_PORT, 0, 100);
	if (!offset_srv) {
		log_send(LOG_T_ERROR, "Could not open listen port for offset");
	}

	command_server_handler_raw_set(offset_srv, handle_offsets);
	command_server_handler_raw_set(console_srv, handle_con_client);
	
	while (1) {
		fd_set fd_rd;
		struct timeval timeout;
		int high;
		struct spg_client *client, **clientp;
		
		timeout.tv_usec = 0;
		timeout.tv_sec = 1;
		
		FD_ZERO(&fd_rd);
		high = fd_listen + 1;
		FD_SET(fd_listen, &fd_rd);

		status_server_fdset_add(stat_srv, &fd_rd, &high);
		command_server_fdset_add(offset_srv, &fd_rd, &high);
		command_server_fdset_add(console_srv, &fd_rd, &high);
		
		for (client = clients; client; client = client->next) {
			if (!client->pending) {
				command_fd_set(client->command, &fd_rd, &high);
			} else {
				try_command(client->command);
			}
		}

		select(high + 1, &fd_rd, NULL, NULL, &timeout);

		if (FD_ISSET(fd_listen, &fd_rd)) {
			new_connection();
		}

		status_server_fdset_handle(stat_srv, &fd_rd);
		command_server_fdset_handle(offset_srv, &fd_rd);
		command_server_fdset_handle(console_srv, &fd_rd);

		for (client = clients, clientp = &clients; client; client = *clientp) {
			if (!client->pending)
				command_handle(client->command, &fd_rd);
			if (client->cleanup) {
				*clientp = client->next;
				log_send(LOG_T_DEBUG, "Removed unknown spg client");
				command_free(client->command);
				free(client);
			} else {
				clientp = &client->next;
			}
		}

		update_status(stat_srv);
	}
}
