/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>


/*
       inputs      outputs
       nr name     nr name

     ----------------------
     |                    |
  ---| 0  in       0  out |----
     |	                  |
     ----------------------
     
 */

struct controller_block_private {
	float *input0;
	float out;
	
	float alpha;
	float t;
	float tau;
};

static void filter_lp_calculate(struct controller_block *lp)
{
	struct controller_block_private *priv = lp->private;
	float alpha = priv->alpha;
	float out = priv->out;
	
	out *= 1.0 - alpha;
	out += *lp->private->input0 * alpha;
	priv->out = out;
}

static struct controller_block_param_list params[] = {
	{ "tau", true },
	{ "t",   true },
	{ NULL },
};

static void param_get(struct controller_block *lp, int param, void *val)
{
	switch (param) {
		case 0:
			*(float*)val = lp->private->tau;
			break;
		case 1:
			*(float*)val = lp->private->t;
			break;
	}
}

static void param_set(struct controller_block *lp, int param, va_list val)
{
	switch (param) {
		case 0:
			lp->private->tau = va_arg(val, double);
			break;
		case 1:
			lp->private->t = va_arg(val, double);
			break;
	}

	if (lp->private->t == 0.0)
		lp->private->t = 1.0;

	lp->private->alpha =
	    lp->private->t / (lp->private->t + lp->private->tau);
}

static struct controller_block_interm_list interms[] = {
	{ "in", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input0) },
	{ NULL }
};

static struct controller_block_outterm_list outterms[] = {
	{ "out", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out) },
	{ NULL }
};

struct controller_block * block_filter_lp_create(char *name)
{
	struct controller_block *lp;
	
	if (!(lp = controller_block_alloc("filter_lp", name, sizeof(struct controller_block_private))))
		return NULL;

	lp->private->out = 0.0;
	lp->private->t = 1.0;
	lp->private->tau = 1.0;
	lp->private->alpha = 0.0;

	if (controller_block_interm_list_init(lp, interms))
		goto err_block;

	if (controller_block_outterm_list_init(lp, outterms))
		goto err_block;

	lp->calculate = filter_lp_calculate;

	if (controller_block_param_list_init(lp, params))
		goto err_block;

	lp->param_get = param_get;
	lp->param_set = param_set;

	controller_block_add(lp);
	return lp;

err_block:
	controller_block_free(lp);
	return NULL;
}
