/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <controller/controller_block.h>

/*
       inputs          outputs
       nr name         nr name

     --------------------------
  ---| 0  in                  |
     |                 0  out |----
  ---| 1  speed               |
     --------------------------

	When speed is within the deadzone 
	(deadzone_min <= speed <= deadzone_max) the static friction is used.
	There is a limit to make sure the output does not goes through zero:
	
		in - friction_static >= 0:
		
			out = in - friction_static

		in - friction_static < 0:
		
			out = 0;
	
	When speed is outside the deadzone (deadzone_min > speed) or
	(deadzone_max < speed) the dynamic friction is used.
	The result can go through zero:
	
			out = in - firction_dynamic

      
 */

struct controller_block_private {
	float *input0;
	float *input1;
	float out;
	
	float deadzone_max;
	float deadzone_min;
	float friction_static;
	float friction_dynamic;
};

static void friction_calculate(struct controller_block *friction)
{
	struct controller_block_private *priv = friction->private;
	float input0 = *priv->input0;
	float input1 = *priv->input1;

	if (input1 <= priv->deadzone_max &&
	    input1 >= priv->deadzone_min) {
		if (abs(input0) > abs(priv->friction_static)) {
			if (input1 >= 0) {
				priv->out =
				    input0 - priv->friction_static;
			} else {
				priv->out =
				    input0 + priv->friction_static;
			}
		} else {
			priv->out = 0.0;
		}
	} else {
		if (input1 >= 0) {
			priv->out = 
			    input0 - priv->friction_dynamic;
		} else {
			priv->out =
			    input0 + priv->friction_dynamic;
		}
	}
}

static struct controller_block_param_list params[] = {
	{ "friction_static",  false },
	{ "friction_dynamic", false },
	{ "deadzone_max",     false },
	{ "deadzone_min",     false },
	{ NULL },
};

static void param_get(struct controller_block *friction, int param, void *val)
{
	struct controller_block_private *priv = friction->private;

	switch (param) {
		case 0:
			*(float*)val = priv->friction_static;
			break;
		case 1:
			*(float*)val = priv->friction_dynamic;
			break;
		case 2:
			*(float*)val = priv->deadzone_max;
			break;
		case 3:
			*(float*)val = priv->deadzone_min;
			break;
	}
}

static void param_set(struct controller_block *friction, int param, va_list val)
{
	struct controller_block_private *priv = friction->private;

	switch (param) {
		case 0:
			priv->friction_static = va_arg(val, double);
			break;
		case 1:
			priv->friction_dynamic = va_arg(val, double);
			break;
		case 2:
			priv->deadzone_max = va_arg(val, double);
			break;
		case 3:
			priv->deadzone_min = va_arg(val, double);
			break;
	}
}

static struct controller_block_interm_list interms[] = {
	{ "in", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input0) },
	{ "speed", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input1) },
	{ NULL }
};

static struct controller_block_outterm_list outterms[] = {
	{ "out", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out) },
	{ NULL }
};

struct controller_block * block_friction_create(char *name)
{
	struct controller_block *friction;
	
	if (!(friction = controller_block_alloc("friction", name, sizeof(struct controller_block_private))))
		return NULL;

	friction->private->out = 0.0;
	friction->private->friction_dynamic = 0.0;
	friction->private->friction_static = 0.0;
	friction->private->deadzone_min = 0.0;
	friction->private->deadzone_max = 0.0;
	
	if (controller_block_interm_list_init(friction, interms))
		goto err_block;

	if (controller_block_outterm_list_init(friction, outterms))
		goto err_block;

	friction->calculate = friction_calculate;

	if (controller_block_param_list_init(friction, params))
		goto err_block;

	friction->param_get = param_get;
	friction->param_set = param_set;

	controller_block_add(friction);
	return friction;

err_block:
	controller_block_free(friction);
	return NULL;
}
