/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <controller/controller_block.h>
#include <controller/controller_bus.h>
#include <log/log.h>

#include <linux/joystick.h>
#include <errno.h>

struct controller_block_private {
	float	*axes;
	bool	*buttons;

	char *device;
	int fd;
	
	struct controller_bus *bus;
};

static void calculate(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	struct js_event event;
	int ret;
	
	if (priv->fd < 0)
		return;
	
	do {
		ret = read (priv->fd, &event, sizeof(struct js_event));
		if (ret == sizeof(struct js_event)) {
			if (event.type & JS_EVENT_BUTTON) {
				priv->buttons[event.number] = event.value;
			}
			if (event.type & JS_EVENT_AXIS) {
				priv->axes[event.number] = event.value / 32768.0;
			}
		}
	} while (ret == sizeof(struct js_event));
	
	if (ret <= 0 && errno != EAGAIN) {
		close(priv->fd);
		priv->fd = -1;
		log_send(LOG_T_ERROR, "%s: Error reading device", block->name);
	}
}

static int joystick_open(char *device)
{
	int fd;
	int ret;
	
	fd = open(device, O_RDONLY);
	if (fd < 0)
		return -1;
	
	ret = ioctl(fd, FIONBIO, &(int){1});
	if (ret)
		goto err_ioctl;

	return fd;

err_ioctl:
	close(fd);
	return -1;
}


static enum controller_bus_state joystick_poll(struct controller_bus *bus)
{
	struct controller_block *block = bus->owner;
	int fd = block->private->fd;
	
	if (fd >= 0)
		return CONTROLLER_BUS_STATE_OK;
	
	fd = joystick_open(block->private->device);
	if (fd < 0)
		return CONTROLLER_BUS_STATE_ERROR;
	
	block->private->fd = fd;
	log_send(LOG_T_WARNING, "%s: re-opened %s", block->name,
	    block->private->device);
	
	return CONTROLLER_BUS_STATE_OK;
}


struct controller_block * block_joystick_create(char *name, va_list ap)
{
	struct controller_block *block;
	int fd;
	char *device;
	int nr_axes = 0, nr_buttons = 0;
	int i, ret;

	device = va_arg(ap, char *);

	fd = joystick_open(device);
	if (fd < 0) {
		log_send(LOG_T_ERROR, "Could not open device %s", device);
		return NULL;
	}
	ret = ioctl(fd, JSIOCGAXES, &nr_axes);
	if (ret)
		goto err_ioctl;
	ret = ioctl(fd, JSIOCGBUTTONS, &nr_buttons);
	if (ret)
		goto err_ioctl;

	log_send(LOG_T_DEBUG, "%s: %d axes, %d buttons", 
	    device, nr_axes, nr_buttons);
	

	if (!(block = controller_block_alloc("joystick", name, sizeof(struct controller_block_private))))
		goto err_block;

	block->private->axes = calloc(sizeof(float), nr_axes);
	if (!block->private->axes)
		goto err_axes;
	block->private->buttons = calloc(sizeof(bool), nr_buttons);
	if (!block->private->buttons)
		goto err_buttons;

	block->private->device = strdup(device);

	block->inputs = 0;
	block->input = NULL;
	
	/* Inputs on the hardware are outputs on this software block */
	block->outputs = nr_axes + nr_buttons;
	block->output = malloc(sizeof(struct controller_block_outterm) * 
	    (nr_axes + nr_buttons));
	if (!block->output)
		goto err_output;
	for (i = 0; i < nr_axes; i++) {
		block->output[i].type = CONTROLLER_BLOCK_TERM_FLOAT;
		block->output[i].value.f = &block->private->axes[i];
		block->output[i].source = block;
		block->output[i].name = malloc(16);
		sprintf(block->output[i].name, "axis%d", i);
	}
	for (i = 0; i < nr_buttons; i++) {
		int outnr = i + nr_axes;
		block->output[outnr].type = CONTROLLER_BLOCK_TERM_BOOL;
		block->output[outnr].value.b = &block->private->buttons[i];
		block->output[outnr].source = block;
		block->output[outnr].name = malloc(16);
		sprintf(block->output[outnr].name, "button%d", i);
	}
	
	block->private->bus = controller_bus_create("joystick", name, block, 0);
	if (!block->private->bus)
		goto err_bus;

	controller_bus_poll_set(block->private->bus, joystick_poll, NULL);

	block->calculate = calculate;

	block->params = 0;
	block->param = NULL;
	block->param_get = NULL;
	block->param_set = NULL;
	
	block->private->fd = fd;

	controller_block_add(block);
	return block;

err_bus:
err_output:
	free(block->private->buttons);
	free(block->private->device);
err_buttons:
	free(block->private->axes);
err_axes:
	free(block->private);
err_block:
	controller_block_free(block);
err_ioctl:
	close(fd);
	return NULL;
}
