/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007, 2011
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <controller/controller_block.h>


/*
       inputs      outputs
       nr name     nr name

     ----------------------
     |                    |
  ---| 0  in       0  out |----
     |	                  |
  ---| 1  enable          |
     |	                  |
     ----------------------

                             min/max
                                :
                            /-------\
  in   /---\ in' /---\ in'' |    -- |   /---\         /---\        out
  -----| - |-----| - |------|   /   |---| + |---------| + |-----------
       \---/     \---/      | --    |   \---/       | \---/       |
         |         |        \-------/     |         |   |         |
	 |         |                      |  /---\  |   |         |
	 |         \-------------------------|Z-1|--/   |         |
         |                                   \---/      |         |
	 |                                              |  /---\  |
	 \-------------------------------------------------|Z-1|--/
                                                           \---/
 */

struct controller_block_private {
	float *input0;
	bool *enable;
	float out;
	
	float der_last;
	
	float min;
	float max;
	float imin;
	float imax;
	
	float mint;
	float maxt;
	float t;
};

static void limit_2nd_calculate(struct controller_block *limit)
{
	struct controller_block_private *priv = limit->private;
	float der_1;
	float der_2;
	float tneeded;
	float ttleft;
	float input0 = *priv->input0;
	float max = priv->max;
	float min = priv->min;
	float der_last = priv->der_last;
	float out = priv->out;
	float imax = priv->imax;
	float imin = priv->imin;
	bool enable = *priv->enable;
	
	der_1 = input0 - out;
	der_2 = der_1 - der_last;

	if (der_2 > max) {
		tneeded = der_last * imax;
		ttleft = 2 * der_1 * imax;
		if (tneeded >= 0 && ttleft >= 0 &&
		    tneeded * tneeded >= ttleft)
			der_2 = min;
		else
			der_2 = max;
	}
	if (der_2 < min) {
		tneeded = der_last * imin;
		ttleft = 2 * der_1 * imin;
		if (tneeded >= 0 && ttleft >= 0 &&
		    tneeded * tneeded >= ttleft)
			der_2 = max;
		else
			der_2 = min;
	}

	der_last += der_2;
	out += der_last;

	if (!enable) {
		out = 0.0;
		der_last = 0.0;
	}
	priv->der_last = der_last;
	priv->out = out;
}

static struct controller_block_param_list params[] = {
	{ "min",   true },
	{ "max",   true },
	{ "t",     true },
	{ "reset", true },
	{ NULL },
};

static void param_get(struct controller_block *limit, int param, void *val)
{
	switch (param) {
		case 0:
			*(float*)val = limit->private->mint;
			break;
		case 1:
			*(float*)val = limit->private->maxt;
			break;
		case 2: *(float*)val = limit->private->t;
			break;
		case 3:
			break;
	}
}

static void param_set(struct controller_block *limit, int param, va_list val)
{
	switch (param) {
		case 0:
			limit->private->mint = va_arg(val, double);
			break;
		case 1:
			limit->private->maxt = va_arg(val, double);
			break;
		case 2:
			limit->private->t = va_arg(val, double);
			break;
		case 3:
			limit->private->der_last = 0;
			limit->private->out = 0;
			break;
	}
	limit->private->min = limit->private->mint 
	    * limit->private->t * limit->private->t;
	limit->private->max = limit->private->maxt
	    * limit->private->t * limit->private->t;
	limit->private->imin = 1.0 / limit->private->min;
	limit->private->imax = 1.0 / limit->private->max;
}

static struct controller_block_interm_list interms[] = {
	{ "in", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input0) },
	{ "enable", CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, enable) },
	{ NULL }
};

static struct controller_block_outterm_list outterms[] = {
	{ "out", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, out) },
	{ NULL }
};

struct controller_block * block_limit_2nd_create(char *name)
{
	struct controller_block *limit;
	
	if (!(limit = controller_block_alloc("limit_2nd", name, sizeof(struct controller_block_private))))
		return NULL;

	limit->private->out = 0.0;
	limit->private->max = 0.0;
	limit->private->min = 0.0;
	limit->private->imax = 1.0;
	limit->private->imin = 1.0;
	limit->private->maxt = 0.0;
	limit->private->mint = 0.0;
	limit->private->t = 1.0;
	limit->private->der_last = 0.0;

	if (controller_block_interm_list_init(limit, interms))
		goto err_block;

	if (controller_block_outterm_list_init(limit, outterms))
		goto err_block;

	limit->calculate = limit_2nd_calculate;

	if (controller_block_param_list_init(limit, params))
		goto err_block;

	limit->param_get = param_get;
	limit->param_set = param_set;

	controller_block_add(limit);
	return limit;

err_block:
	controller_block_free(limit);
	return NULL;
}
