/*
	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2013, 2014

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdlib.h>
#include <string.h>

#include <log/log.h>
#include <controller/controller_bus.h>


static struct controller_bus *bus_list = NULL;

struct controller_bus *controller_bus_create(char *type, char *name,
    struct controller_block *owner,
    size_t private_size)
{
	struct controller_bus **entry;
	
	for (entry = &bus_list; *entry; entry = &(*entry)->next) {
		if (!strcmp((*entry)->name, name)) {
			log_send(LOG_T_ERROR, 
			    "Bus named '%s' already exists. (type= %s)",
			    name, (*entry)->type);
			return NULL;
		}
	}
	*entry = calloc(1, sizeof(struct controller_bus));
	if (!*entry)
		return NULL;
	if (private_size) {
		(*entry)->private = calloc(1, private_size);
		if (!(*entry)->private) {
			goto err_private;
		}
	}
	
	(*entry)->name = strdup(name);
	(*entry)->type = strdup(type);
	(*entry)->owner = owner;
	(*entry)->input_block = owner;
	(*entry)->output_block = owner;
	(*entry)->state = CONTROLLER_BUS_STATE_OK;
	
	log_send(LOG_T_DEBUG, "New bus '%s' created of type '%s'",
	    name, type);

	return *entry;

err_private:
	free(*entry);
	*entry = NULL;
	return NULL;
}

int controller_bus_input_set(struct controller_bus *bus,
    struct controller_block *block)
{
	bus->input_block = block;
	
	log_send(LOG_T_DEBUG, "Block '%s' handles input for bus '%s'",
	    block->name, bus->name);

	return 0;
}

int controller_bus_output_set(struct controller_bus *bus,
    struct controller_block *block)
{
	bus->output_block = block;

	log_send(LOG_T_DEBUG, "Block '%s' handles output for bus '%s'",
	    block->name, bus->name);
	
	return 0;
}

int controller_bus_poll_set(struct controller_bus *bus,
    enum controller_bus_state (*poll)(struct controller_bus *bus),
    int (*recover)(struct controller_bus *bus))
{
	bus->poll = poll;
	bus->recover = recover;
	
	return 0;
}

struct controller_bus *controller_bus_find(char *name)
{
	struct controller_bus *entry;
	
	for (entry = bus_list; entry; entry = entry->next)
		if (!strcmp(entry->name, name))
			return entry;

	return NULL;
}

struct controller_bus *controller_bus_by_owner(struct controller_block *owner)
{
	struct controller_bus *entry;
	
	for (entry = bus_list; entry; entry = entry->next)
		if (entry->owner == owner)
			return entry;
	
	return NULL;
}

struct controller_bus_client *controller_bus_add_input_client(
    struct controller_bus *bus,
    struct controller_block *client)
{
	struct controller_bus_client *ic;
	int i;
	
	for (i = 0; i < bus->input_clients; i++) {
		if (bus->input_client[i].block == client)
			return &bus->input_client[i];
	}
	
	ic = realloc(bus->input_client,
	    sizeof(struct controller_bus_client) * (bus->input_clients + 1));
	if (!ic)
		return NULL;
	
	bus->input_client = ic;
	bus->input_client[bus->input_clients].block = client;
	bus->input_client[bus->input_clients].callback = NULL;
	bus->input_client[bus->output_clients].poll = NULL;
	bus->input_client[bus->output_clients].recover = NULL;
	bus->input_client[bus->input_clients].offset = 0;
	bus->input_client[bus->output_clients].state = CONTROLLER_BUS_STATE_OK;
	bus->input_clients++;

	log_send(LOG_T_DEBUG, "Block '%s' is an input client of bus '%s'",
	    client->name, bus->name);
	
	return &bus->input_client[bus->input_clients-1];
}

struct controller_bus_client *controller_bus_add_output_client(
    struct controller_bus *bus,
    struct controller_block *client)
{
	struct controller_bus_client *oc;
	int i;
	
	for (i = 0; i < bus->output_clients; i++) {
		if (bus->output_client[i].block == client)
			return &bus->output_client[i];
	}
	
	oc = realloc(bus->output_client,
	    sizeof(struct controller_bus_client) * (bus->output_clients + 1));
	if (!oc)
		return NULL;

	bus->output_client = oc;
	bus->output_client[bus->output_clients].block = client;
	bus->output_client[bus->output_clients].callback = NULL;
	bus->output_client[bus->output_clients].poll = NULL;
	bus->output_client[bus->output_clients].recover = NULL;
	bus->output_client[bus->output_clients].offset = 0;
	bus->output_client[bus->output_clients].state = CONTROLLER_BUS_STATE_OK;
	bus->output_clients++;
	
	log_send(LOG_T_DEBUG, "Block '%s' is an output client of bus '%s'",
	    client->name, bus->name);

	return &bus->output_client[bus->output_clients-1];
}

bool controller_bus_linked(struct controller_block *src, 
    struct controller_block *dst)
{
	struct controller_bus *entry;
	
	for (entry = bus_list; entry; entry = entry->next) {
		int i;
		if (entry->output_block == dst) {
			for (i = 0; i < entry->output_clients; i++)
				if (entry->output_client[i].block == src) {
					return true;
				}
		}
		if (entry->input_block == src) {
			for (i = 0; i < entry->input_clients; i++)
				if (entry->input_client[i].block == dst) {
					return true;
				}
		}
	}

	return false;
}

struct controller_bus *controller_bus_linked_bus(struct controller_block *src, 
    struct controller_block *dst)
{
	struct controller_bus *entry;
	
	for (entry = bus_list; entry; entry = entry->next) {
		int i;
		if (entry->output_block == dst) {
			for (i = 0; i < entry->output_clients; i++)
				if (entry->output_client[i].block == src) {
					return entry;
				}
		}
		if (entry->input_block == src) {
			for (i = 0; i < entry->input_clients; i++)
				if (entry->input_client[i].block == dst) {
					return entry;
				}
		}
	}

	return NULL;
}

int controller_bus_poll_states(int *oks, int *errors, int *recoverables)
{
	struct controller_bus *entry;
	int error = 0;
	int recoverable = 0;
	int ok = 0;
	
	for (entry = bus_list; entry; entry = entry->next) {
		int i;
		int entry_error = 0;
		int entry_recoverable = 0;

		if (entry->poll) {
			enum controller_bus_state newstate;
			
			newstate = entry->poll(entry);
				
			if (newstate == CONTROLLER_BUS_STATE_ERROR) {
				entry_error++;
				error++;
			}
			else if (newstate == CONTROLLER_BUS_STATE_RECOVERABLE) {
				entry_recoverable++;
				recoverable++;
			}
			else if (newstate == CONTROLLER_BUS_STATE_OK)
				ok++;
				
			if (newstate != entry->state) {
				log_send(LOG_T_WARNING, 
				    "%s: changed state from %s to %s",
				    entry->name, 
				    controller_bus_state_2_str(entry->state),
				    controller_bus_state_2_str(newstate)); 
			}
			entry->state = newstate;
		}

		for (i = 0; i < entry->input_clients; i++) {
			if (entry->input_client[i].poll) {
				enum controller_bus_state newstate;
				
				newstate = entry->input_client[i].poll(entry->input_client+i);
				
				if (newstate == CONTROLLER_BUS_STATE_ERROR) {
					entry_error++;
					error++;
				} else if (newstate == CONTROLLER_BUS_STATE_RECOVERABLE) {
					entry_recoverable++;
					recoverable++;
				}
				else if (newstate == CONTROLLER_BUS_STATE_OK)
					ok++;
				
				if (newstate != entry->input_client[i].state) {
					log_send(LOG_T_WARNING, 
					    "%s: Input client %s changed state from %s to %s",
					    entry->name, entry->input_client[i].block->name,
					    controller_bus_state_2_str(entry->input_client[i].state),
					    controller_bus_state_2_str(newstate)); 
				}
				entry->input_client[i].state = newstate;
			}
		}
		for (i = 0; i < entry->output_clients; i++) {
			if (entry->output_client[i].poll) {
				enum controller_bus_state newstate;
				
				newstate = entry->output_client[i].poll(entry->output_client+i);
				
				if (newstate == CONTROLLER_BUS_STATE_ERROR) {
					entry_error++;
					error++;
				}
				else if (newstate == CONTROLLER_BUS_STATE_RECOVERABLE) {
					entry_recoverable++;
					recoverable++;
				}
				else if (newstate == CONTROLLER_BUS_STATE_OK)
					ok++;
				
				if (newstate != entry->output_client[i].state) {
					log_send(LOG_T_WARNING, 
					    "%s: output client %s changed state from %s to %s",
					    entry->name, entry->output_client[i].block->name, 
					    controller_bus_state_2_str(entry->output_client[i].state),
					    controller_bus_state_2_str(newstate));
				}
				entry->output_client[i].state = newstate;
			}
		}
		
		if (entry_error)
			entry->state = CONTROLLER_BUS_STATE_ERROR;
		if (entry_recoverable)
			entry->state = CONTROLLER_BUS_STATE_RECOVERABLE;
	}
	
	if (oks)
		*oks = ok;
	if (errors)
		*errors = error;
	if (recoverables)
		*recoverables = recoverable;
	return 0;
}

int controller_bus_recover(void)
{
	struct controller_bus *entry;

	for (entry = bus_list; entry; entry = entry->next) {
		int i;

		if (entry->state == CONTROLLER_BUS_STATE_RECOVERABLE &&
		    entry->recover) {
			log_send(LOG_T_INFO, 
			    "%s: Attempt to recover", entry->name);
				
			entry->recover(entry);
		}

		for (i = 0; i < entry->input_clients; i++) {
			if (entry->input_client[i].state == CONTROLLER_BUS_STATE_RECOVERABLE &&
			    entry->input_client[i].recover) {
				log_send(LOG_T_INFO, 
				    "%s: Attempt to recover input client %s",
				    entry->name, entry->input_client[i].block->name);
				
				entry->input_client[i].recover(entry->input_client+i);
			}
		}
		for (i = 0; i < entry->output_clients; i++) {
			if (entry->output_client[i].state == CONTROLLER_BUS_STATE_RECOVERABLE &&
			    entry->output_client[i].recover) {
				log_send(LOG_T_INFO, 
				    "%s: Attempt to recover output client %s",
				    entry->name, entry->output_client[i].block->name);
				
				entry->output_client[i].recover(entry->output_client+i);
			}
		}
	}

	return 0;
}

char *controller_bus_state_2_str(enum controller_bus_state state)
{
	switch (state) {
		case CONTROLLER_BUS_STATE_OK:
			return "OK";
		case CONTROLLER_BUS_STATE_RECOVERABLE:
			return "RECOVERABLE";
		case CONTROLLER_BUS_STATE_ERROR:
			return "ERROR";
	}
	
	return "INVALID";
}
