/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007, 2008
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007, 2008

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>
#include <dt_azimuth/dt_az_safety.h>
#include <log/log.h>

/*
       inputs          outputs
       nr name         nr name

     ---------------------------------
     |                               |
  ---| 0  speed_in     0  speed_out  |----
     |	                             |
  ---| 1  position_in  1  safe_out   |----
     |	                             |
  ---| 2  torque_in    2  torque_out |----
     |	                             |
  ---| 3  safety_in_positive         |
     |                               |
  ---| 4  safety_in_negative         |
     |                               |
  ---| 5  enable                     |
     |                               |
     ---------------------------------

	This is not a generic block.
	It implements the safety checks specific for the dt azimuth
	control loop.

	Checks:	
		+ Enabled
		+ Direction limit if position beyond max/min
		+ Speed limit if within safe-zone.

       min           safe_zone_min                safe_zone_max           max
        |------------------|----------------------------|------------------|
        safe_zone_min_speed       normal operation      safe_zone_max_speed

 */

struct controller_block_private {
	float *speed_in;
	float *position_in;
	float *torque_in;
	bool *safety_in_positive;
	bool *safety_in_negative;
	float speed_out;
	bool safe_out;
	float torque_out;
	bool *enable_in;

	float position_min;
	float position_max;
	float emergency_torque;
	
	float safe_zone_min;
	float safe_zone_max;
	float safe_zone_min_speed;
	float safe_zone_max_speed;

	/* keep log state to prevent flooding */
	bool err_max;
	bool err_min;
	bool warn_safe_max;
	bool warn_safe_min;
	bool err_hwsafe_positive;
	bool err_hwsafe_negative;

	bool recover;
};

static void dt_az_safety_calculate(struct controller_block *safety)
{
	float out;
	bool safe = true;
	bool enabled = *safety->private->enable_in;

	out = *safety->private->speed_in;

	if (*safety->private->position_in >= safety->private->position_max &&
	    out > 0.0) {
		out = 0.0;
		safe = false;
		if (!safety->private->err_max) {
			safety->private->err_max = true;
			log_send(LOG_T_ERROR, "Azimuth position above maximum %e > %e",
			    *safety->private->position_in,
			    safety->private->position_max);
		}
	} else {
		safety->private->err_max = false;
	}
	if (*safety->private->position_in >= safety->private->safe_zone_max &&
	    out > safety->private->safe_zone_max_speed) {
		out = safety->private->safe_zone_max_speed;
		if (!safety->private->warn_safe_max) {
			safety->private->warn_safe_max = true;
			log_send(LOG_T_WARNING, "Azimuth position in safe zone");
		}
	} else {
		safety->private->warn_safe_max = false;
	}

	if (*safety->private->position_in <= safety->private->position_min &&
	    out < 0.0) {
		out = 0.0;
		safe = false;
		if (!safety->private->err_min) {
			safety->private->err_min = true;
			log_send(LOG_T_ERROR, "Azimuth position under minimum %e < %e",
			    *safety->private->position_in,
			    safety->private->position_min);
		}
	} else {
		safety->private->err_min = false;
	}
	if (*safety->private->position_in <= safety->private->safe_zone_min &&
	    out < safety->private->safe_zone_min_speed) {
		out = safety->private->safe_zone_min_speed;
		if (!safety->private->warn_safe_min) {
			safety->private->warn_safe_min = true;
			log_send(LOG_T_ERROR, "Azimuth position in safe zone");
		}
	} else {
		safety->private->warn_safe_min = false;
	}
	
	if (*safety->private->safety_in_positive == 0) {
		safe = false;
		if (!safety->private->err_hwsafe_positive) {
			safety->private->err_hwsafe_positive = true;
			log_send(LOG_T_ERROR, "Azimuth HW positive safety switch is open.");
		}
	} else {
		if (safety->private->err_hwsafe_positive) {
			safety->private->err_hwsafe_positive = false;
			log_send(LOG_T_INFO, "Azimuth HW positive safety switch is closed.");
		}
	}
	if (*safety->private->safety_in_negative == 0) {
		safe = false;
		if (!safety->private->err_hwsafe_negative) {
			safety->private->err_hwsafe_negative = true;
			log_send(LOG_T_ERROR, "Azimuth HW negative safety switch is open.");
		}
	} else {
		if (safety->private->err_hwsafe_negative) {
			safety->private->err_hwsafe_negative = false;
			log_send(LOG_T_INFO, "Azimuth HW negative safety switch is closed.");
		}
	}

	if (!safe) {
		enabled = false;
	}

	if (!enabled && !safety->private->recover) {
		out = 0.0;
		safety->private->torque_out = safety->private->emergency_torque;
	} else {
		safety->private->torque_out = *safety->private->torque_in;
	}

	safety->private->safe_out = safe;
	safety->private->speed_out = out;
}

static struct controller_block_param_list params[] = {
	{ "position_min",        false },
	{ "position_max",        false },
	{ "safe_zone_min",       false },
	{ "safe_zone_max",       false },
	{ "safe_zone_min_speed", false },
	{ "safe_zone_max_speed", false },
	{ "emergency_torque",    false },
	{ "recover",             false },
	{ NULL },
};

static void param_get(struct controller_block *safety, int param, void *val)
{
	switch (param) {
		case 0:
			*(float*)val = safety->private->position_min;
			break;
		case 1:
			*(float*)val = safety->private->position_max;
			break;
		case 2:
			*(float*)val = safety->private->safe_zone_min;
			break;
		case 3:
			*(float*)val = safety->private->safe_zone_max;
			break;
		case 4:
			*(float*)val = safety->private->safe_zone_min_speed;
			break;
		case 5:
			*(float*)val = safety->private->safe_zone_max_speed;
			break;
		case 6:
			*(float*)val = safety->private->emergency_torque;
			break;
		case 7:
			*(int*)val = safety->private->recover;
			break;
	}
}

static void param_set(struct controller_block *safety, int param, va_list val)
{
	switch (param) {
		case 0:
			safety->private->position_min = va_arg(val, double);
			break;
		case 1:
			safety->private->position_max = va_arg(val, double);
			break;
		case 2:
			safety->private->safe_zone_min = va_arg(val, double);
			break;
		case 3:
			safety->private->safe_zone_max = va_arg(val, double);
			break;
		case 4:
			safety->private->safe_zone_min_speed = va_arg(val, double);
			break;
		case 5:
			safety->private->safe_zone_max_speed = va_arg(val, double);
			break;
		case 6:
			safety->private->emergency_torque = va_arg(val, double);
			break;
		case 7: {
			bool recover = va_arg(val, int);
		
			if (recover && !safety->private->recover) {
				log_send(LOG_T_WARNING,
				    "%s: Switching to recover mode, beware!",
				    safety->name);
			}
			if (!recover && safety->private->recover) {
				log_send(LOG_T_INFO,
				    "%s: Switching back to normal mode",
				    safety->name);
			}
		
			safety->private->recover = recover;
		
			break;
		}
	}
}

static struct controller_block_interm_list interms[] = {
	{ "speed_in",           CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, speed_in)           },
	{ "position_in",        CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, position_in)        },
	{ "torque_in",          CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, torque_in)          },
	{ "safety_in_positive", CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private,  safety_in_positive) },
	{ "safety_in_negative", CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private,  safety_in_negative) },
	{ "enable",             CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private,  enable_in)          },
	{ NULL }
};

static struct controller_block_outterm_list outterms[] = {
	{ "speed_out",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, speed_out)  },
	{ "safe_out",   CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, safe_out)   },
	{ "torque_out", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, torque_out) },
	{ NULL }
};

struct controller_block * block_dt_az_safety_create(char *name)
{
	struct controller_block *safety;
	
	safety = malloc(sizeof(struct controller_block));
	if (!safety)
		return NULL;

	safety->type = "dt_az_safety";
	safety->name = malloc(strlen(name)+1);
	if (!safety->name)
		goto err_safety;
	strcpy(safety->name, name);

	safety->private = malloc(sizeof(struct controller_block_private));
	if (!safety->private)
		goto err_name;
	safety->private->speed_out = 0.0;
	safety->private->safe_out = 0.0;
	safety->private->position_min = 0.0;
	safety->private->position_max = 0.0;
	safety->private->safe_zone_min = 0.0;
	safety->private->safe_zone_max = 0.0;
	safety->private->safe_zone_min_speed = 0.0;
	safety->private->safe_zone_max_speed = 0.0;
	safety->private->emergency_torque = 0.0;

	safety->private->err_max = false;
	safety->private->err_min = false;
	safety->private->warn_safe_max = false;
	safety->private->warn_safe_min = false;
	safety->private->err_hwsafe_positive = false;
	safety->private->err_hwsafe_negative = false;
	
	safety->private->recover = false;

	if (controller_block_interm_list_init(safety, interms))
		goto err_private;

	if (controller_block_outterm_list_init(safety, outterms))
		goto err_input;


	safety->calculate = dt_az_safety_calculate;

	if (controller_block_param_list_init(safety, params))
		goto err_output;

	safety->param_get = param_get;
	safety->param_set = param_set;

	controller_block_add(safety);
	return safety;

err_output:
	free(safety->output);
err_input:
	free(safety->input);
err_private:
	free(safety->private);
err_name:
	free(safety->name);
err_safety:
	free(safety);
	return NULL;
}
