/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2007 - 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2007 - 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
/*
	dt_ctr.c
	
	Main loop for the DT controller.
	Here the controller network is initialized.
	(By loading a .ctrl file supplied on the command line).
	After initialization the sample thread is started and this function's
	work is done.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <signal.h>

#include "dynarg.h"

#include <controller/controller_bus.h>
#include <controller/controller_block.h>
#include <controller/controller_trace.h>
#include <controller/controller_sample.h>
#include <controller/controller_dumpdot.h>
#include <controller/controller_load.h>
#include <shell/shell.h>
#include <ec/ec.h>
#include <log/log.h>

#include <dt_port_numbers.h>



int main(int argc, char **argv)
{
	char *ctrl_filename;
	char *dot_filename;
	sigset_t sigset;

	sigemptyset (&sigset);
	sigaddset(&sigset, SIGALRM);
	sigprocmask(SIG_BLOCK, &sigset, NULL);

	log_server_start(CTRL_LOG_PORT, LOG_T_DEBUG, LOG_T_INFO);

	controller_block_param_init();
	
	if (argc < 2) {
		printf("Usage: %s <controllerfile>\n", argv[0]);
		printf("\n");
		printf("E.g.: %s dt_ctrl.ctrl\n", argv[0]);
		goto err_init;
	}
	
	ctrl_filename = argv[1];
	
	/* Create and link blocks */
	log_send(LOG_T_DEBUG, "Going to load controller file '%s'", 
	    ctrl_filename);
	if (controller_load(ctrl_filename)) {
		log_send(LOG_T_ERROR, "Could not load controller file");
		goto err_init;
	}

	controller_trace_server_start(CTRL_TRACE_PORT, 100);

	/* Start command shell */
	
	controller_sample_shell_add();
	controller_load_shell_add();

	shell_intro = "Dwingeloo Telescope controller\n";
	shell_prompt = "DT>";
	
	shell_server_start(CTRL_SHELL_PORT, 10);

	asprintf(&dot_filename, "%s.dot", ctrl_filename);
	controller_dumpdot(dot_filename);
	free(dot_filename);

	/* Start 'sample' */

	if (controller_block_sample_init()) {
		goto err_init;
	}

	controller_sample_start();

	log_send(LOG_T_DEBUG, "Entering state polling loop");
	int oks_p = 0, errors_p = 0, recoverables_p = 0;
	while (1) {
		int oks, errors, recoverables;
		
		controller_bus_poll_states(&oks, &errors, &recoverables);
		if (oks != oks_p) {
			log_send(LOG_T_WARNING, 
			    "Number of OK blocks changed: %d -> %d",
			    oks_p, oks);
			oks_p = oks;
		}
		if (errors != errors_p) {
			log_send(LOG_T_WARNING, 
			    "Number of ERROR blocks changed: %d -> %d",
			    errors_p, errors);
			errors_p = errors;
		}
		if (recoverables != recoverables_p) {
			log_send(LOG_T_WARNING, 
			    "Number of RECOVERABLE blocks changed: %d -> %d",
			    recoverables_p, recoverables);
			recoverables_p = recoverables;
		}
		
		if (recoverables) {
			controller_bus_recover();
		}

		sleep(1);
	}

	return 0;

err_init:
	log_server_flush();
	return 1;
}
