/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2011, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2011

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <controller/controller_block.h>
#include <controller/controller_bus.h>

#include <ec/block_beckhoff_el2502.h>
#include <ec/ec.h>
#include <ec/esc.h>
#include <ec/esc_esi.h>
#include <ec/esc_coe.h>
#include <ec/esc_id.h>

#define EL2502_COE_RXPDO_ASSIGN		0x1c12
#define EL2502_COE_RXPDO_ASSIGN_O1_PWM		0x1600	/* only PWM */
#define EL2502_COE_RXPDO_ASSIGN_O2_PWM		0x1601	/* only PWM */
#define EL2502_COE_RXPDO_ASSIGN_O1_PWM_PER	0x1602	/* PWM and period */
#define EL2502_COE_RXPDO_ASSIGN_O2_PWM_PER	0x1603	/* PWM and period */

#define EL2502_PERIOD_20	0.05
#define EL2502_PERIOD_20000	0.00005

struct controller_block_private {
	float	*output1;
	float	*period1;
	float	*output2;
	float	*period2;
	
	unsigned char *tx_buffer;
	struct canopen_dev *candev;
};

static void calculate(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	float output1, output2, period1, period2;
	union {
		uint8_t  output_c[8];
		uint16_t output_u[4];
	} output;
	
	output1 = fmax(1.0, fabs(*priv->output1)) * 65535.0;
	output2 = fmax(1.0, fabs(*priv->output2)) * 65535.0;

	period1 = fmax(EL2502_PERIOD_20000, fmin(EL2502_PERIOD_20, fabs(*priv->period1))) * 1000000.0;
	period2 = fmax(EL2502_PERIOD_20000, fmin(EL2502_PERIOD_20, fabs(*priv->period2))) * 1000000.0;

	output.output_u[0] = (uint16_t)output1;
	output.output_u[1] = (uint16_t)period1;
	output.output_u[2] = (uint16_t)output2;
	output.output_u[3] = (uint16_t)period2;

	memcpy(priv->tx_buffer, output.output_c, sizeof(output.output_c));
}


static void callback_tx(struct controller_block *block, void *data)
{
	block->private->tx_buffer = data;
}

struct beckhoff_el2502_type {
	char *name;
	uint32_t productcode;
};

static struct beckhoff_el2502_type devlist[] = {
	{ "EL2502", ESC_ESI_PRODUCTCODE_BECKHOFF_EL2502 },
};

static struct controller_block_interm_list interms[] = {
	{ "output1",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, output1)  },
	{ "period1",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, period1)  },
	{ "output2",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, output2)  },
	{ "period2",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, period2)  },
	{ NULL },
};

static int pre_operational_hook(struct esc_device *esc)
{
	struct controller_block *block = esc->block;
	struct canopen_dev *canopen_dev;
	ssize_t r;
	uint16_t pdo_assign;
		
	if (!block->private->candev)
		block->private->candev = esc_coe_create(esc);
	canopen_dev = block->private->candev;

	/* Set the rxpdo such that we can set both pwm and period values */

	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 1, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 1: 0x%04x\n", le16toh(pdo_assign));
	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 2, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 2: 0x%04x\n", le16toh(pdo_assign));

	pdo_assign = htole16(EL2502_COE_RXPDO_ASSIGN_O1_PWM_PER);
	r = canopen_write_param(canopen_dev,
	    EL2502_COE_RXPDO_ASSIGN, 1, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from write, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	pdo_assign = htole16(EL2502_COE_RXPDO_ASSIGN_O2_PWM_PER);
	r = canopen_write_param(canopen_dev,
	    EL2502_COE_RXPDO_ASSIGN, 2, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from write, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}

	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 1, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 1: 0x%04x\n", le16toh(pdo_assign));
	r = canopen_read_param(canopen_dev, 
	    EL2502_COE_RXPDO_ASSIGN, 2, &pdo_assign, sizeof(pdo_assign));
	if (r != sizeof(pdo_assign)) {
		printf("Expected %zd from read, but got %zd!\n",
		    sizeof(pdo_assign), r);
	}
	printf("RxPDO assign for output 2: 0x%04x\n", le16toh(pdo_assign));

	return 0;
}


struct controller_block * block_beckhoff_el2502_create(char *name, va_list ap)
{
	struct controller_block *block;
	int i;
	int dev = -1;
	char *eltype;
	int devno;
	char *busname;
	struct esc_device *esc;
		
	busname = va_arg(ap, char *);
	devno = va_arg(ap, int);
	eltype = va_arg(ap, char *);
	uint32_t vendor, product;
	
	for (i = 0; i < sizeof(devlist)/sizeof(struct beckhoff_el2502_type); i++) {
		if (!strcmp(eltype, devlist[i].name)) {
			dev = i;
			break;
		}
	}
	if (dev < 0)
		return NULL;

	esc = esc_device_create(busname, devno, name);
	
	vendor = esc_esi_vendorid_get(esc);
	product = esc_esi_productcode_get(esc);
	if (vendor != ESC_ESI_VENDORID_BECKHOFF ||
	    product != devlist[dev].productcode) {
		printf("Device not found\n");
		return NULL;
	}

	block = controller_block_alloc("beckhoff_el2502", name,
	    sizeof(struct controller_block_private));
	if (!block)
		return NULL;

	esc_esi_device_fill(esc);
	esc->block = block;
	esc->block_tx = block;
	esc->block_rx = block;
	esc->callback_tx = callback_tx;
	esc->pre_operational_hook = pre_operational_hook;
	
	esc_device_initialize_operational(esc);
	
	/* Outputs on the hardware are inputs on this software block */
	if (controller_block_interm_list_init(block, interms))
		return NULL;	
	
	block->calculate = calculate;

	controller_block_add(block);
	return block;
}
