/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <controller/controller_block.h>

#include <ec/ec.h>
#include <ec/esc.h>
#include <ec/esc_esi.h>
#include <ec/esc_coe.h>
#include <ec/esc_id.h>
#include <log/log.h>

struct controller_block_private {
	float	 converted_value;
	uint32_t value;
	
	double 	gain;
	double 	offset;
	
	unsigned char *rx_buffer;
	
	struct esc_device *dev;
	struct canopen_dev *candev;
};

static void calculate(struct controller_block *block)
{
	struct controller_block_private *priv = block->private;
	struct {
		uint8_t	 status;
		uint32_t value;
	} __packed *rxdata;
	uint32_t value;
	
	rxdata = (void*)priv->rx_buffer;
	
	value = le32toh(rxdata->value);
	priv->value = value;
	priv->converted_value = (value + priv->gain) / priv->offset;
}

static void callback_rx(struct controller_block *block, void *data)
{
	block->private->rx_buffer = data;
}

static struct controller_block_outterm_list outterms[] = {
	{ "converted_value", CONTROLLER_BLOCK_TERM_FLOAT,  offsetof(struct controller_block_private, converted_value)  },
	{ "value",           CONTROLLER_BLOCK_TERM_UINT32, offsetof(struct controller_block_private, value) },
	{ NULL },
};

static int pre_operational_hook(struct esc_device *esc)
{
	struct controller_block *block = esc->block;

	if (!block->private->candev)
		block->private->candev = esc_coe_create(esc);

	return 0;
}

struct controller_block * block_beckhoff_el5001_create(char *name, va_list ap)
{
	struct controller_block *block;
	int devno;
	char *busname;
	uint32_t vendor, product;
	struct esc_device *esc;

	busname = va_arg(ap, char *);
	devno = va_arg(ap, int);
	
	esc = esc_device_create(busname, devno, name);
	
	vendor = esc_esi_vendorid_get(esc);
	product = esc_esi_productcode_get(esc);
	if (vendor != ESC_ESI_VENDORID_BECKHOFF ||
	    product != ESC_ESI_PRODUCTCODE_BECKHOFF_EL5001) {
		log_send(LOG_T_ERROR, "%s: Device not found\n", name);
		return NULL;
	}
	
	block = controller_block_alloc("beckhoff_el5001", name,
	    sizeof(struct controller_block_private));
	if (!block)
		return NULL;

	block->private->dev = esc;
	esc_esi_device_fill(esc);
	esc->block = block;
	esc->block_rx = block;
	esc->callback_rx = callback_rx;
	esc->pre_operational_hook = pre_operational_hook;

	esc_device_initialize_operational(esc);
	

	/* Inputs on the hardware are outputs on this software block */
	if (controller_block_outterm_list_init(block, outterms))
		return NULL;

	block->calculate = calculate;
	
	block->private->gain = 1.0;
	block->private->offset = 0.0;
	
	controller_block_add(block);
	return block;
}
