/*
 * Copyright (c) 2012,
 *      Daan Vreeken <Daan @ Vitsch . nl> - Vitsch Electronics
 *      All rights reserved.
 * Copyright Jeroen Vreeken (jeroen@vreeken.net), 2013
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <controller/controller_block.h>

#include <ec/ec.h>
#include <ec/esc.h>
#include <ec/esc_esi.h>
#include <ec/esc_coe.h>
#include <ec/esc_id.h>
#include <log/log.h>
#include <ec/esc_device.h>


struct controller_block_private {
	unsigned char		*rx_buffer;
	unsigned char		*tx_buffer;

	struct canopen_dev	*can_dev;
	int			fail_cnt;
	
	int32_t			unwound_cnt;
	int32_t			pos_out;
	int32_t			calib_pos;
	uint16_t		last_cnt;
	bool			state_ok;
	
	bool			*enable;
	bool			*calib;
	int16_t			*speed;
	
	struct esc_device	*esc;
	struct canopen_dev	*candev;
};

struct rxpdo_t {
	// PDO 0x1600
	uint8_t			enc_ctrl[4];
	// PDO 0x1602
	uint8_t			stm_ctrl[2];
	// PDO 0x1604
	int16_t			stm_velocity;
} __packed;

struct txpdo_t {
	// PDO 0x1a00
	uint8_t			status_compact[2];
	uint16_t		counter_val;
	uint16_t		latch_val;
	// PDO 0x1a03
	int8_t			full_status[2];
} __packed;

#define FULL_STATUS_FLAG_READY	(1 << 1)

#define MAX_FAIL_CNT		50


// een paar definities voor de leesbaarheid van de code
#define OUTPUT(name)		(priv->name)
#define INPUT(name)		(*priv->name)
#define VAR(name)		(priv->name)


// see:
// http://infosys.beckhoff.com/espanol.php?content=../content/1034/el70x1/html/bt_el7031_objectdescription.htm&id=7476

#define REG_ACTUAL_POS		0x6020, 0x11

#define REG_OUTP_ENABLE		0x7010, 1
#define REG_OUTP_RESET		0x7010, 2
#define REG_OUTP_REDUCE_TORQUE	0x7010, 3
#define REG_OUTP_POSITION	0x7010, 0x11
#define REG_OUTP_VELOCITY	0x7010, 0x21

#define REG_MAX_CURRENT		0x8010, 1	// *1mA
#define REG_REDUCED_CURRENT	0x8010, 2	// *1mA
#define REG_NOMINAL_VOLTAGE	0x8010, 3	// *1mV, max=24000
#define REG_MOT_COIL_RESISTANCE	0x8010, 4	// *.01 Ohm
#define REG_MOT_EMF		0x8010, 5	// ? 1mV / 1000 digits
#define REG_MOT_FULL_STEPS	0x8010, 6	// steps / revolution
#define REG_MAX_START_FREQ	0x8010, 9
#define REG_DRIVE_ON_DELAY	0x8010, 0x10
#define REG_DRIVE_OFF_DELAY	0x8010, 0x11

#define REG_OPERATING_MODE	0x8012, 1
#define OPMODE_DIRECT_VELOCITY			1

#define REG_ACCEL_POS		0x8020, 0x03
#define REG_ACCEL_NEG		0x8020, 0x04
#define REG_DECEL_POS		0x8020, 0x05
#define REG_DECEL_NEG		0x8020, 0x06
#define REG_CALIB_POS		0x8020, 0x08

#define REG_STATUS_WORD		0x9010, 1

// at this moment, the motor's specs are hard-coded for the Nanotech
// ST4118L0804 :
#define MOTOR_VOLTAGE		7400		// mV
#define MOTOR_MAX_CURRENT	800		// mA (avg=0.8A. should be OK)
#define MOTOR_REDUCED_CURRENT	300		// mA
#define MOTOR_COIL_RESISTANCE	9300		// mOhm
#define MOTOR_EMF		0		// ? PDF says '16 koeien per wei bij 300 U/min'
#define MOTOR_STEPS_PER_360DEG	200



static void calculate(struct controller_block *block)
{
	struct controller_block_private	*priv;
	struct rxpdo_t			*rxpdo;
	struct txpdo_t			*txpdo;
	uint16_t			curr_cnt;
	int32_t				unwound_cnt;
	uint8_t				flags = 0;


	priv = block->private;
	
	rxpdo = (struct rxpdo_t *)priv->tx_buffer;
	txpdo = (struct txpdo_t *)priv->rx_buffer;

	if (0) {
		int		cnt;
		uint8_t		*ptr;
		
		// default TxPDO's:
		// 1a00 followed by 1a03
		// 1a00 is 6 bytes in length
		// 1a03 is 2 bytes in length
		
		printf("TxPDO data received:");
		ptr = priv->rx_buffer;
		for (cnt = 0; cnt < 8; cnt++) {
			printf(" %d:%02x", cnt, ptr[cnt]);
		}
		printf("\n");
	}

	//// handle data in TxPDO
	//

	// check 'ready' flag
	if (txpdo->full_status[0] & FULL_STATUS_FLAG_READY) {
		priv->state_ok = true;
		priv->fail_cnt = 0;
		//printf("mot ready\n");
	} else {
		priv->state_ok = false;
		if (INPUT(enable) != true)
			priv->fail_cnt = 0;
		
		if (priv->fail_cnt++ > MAX_FAIL_CNT) {
			// TODO: this is a kludge to detect that the
			// EL7031 has stalled.. can we get out of this
			// state in a non-blocking way??
//			el7031_unhang(priv);
//			priv->state_ok = true;
//			priv->fail_cnt = 0;
		}
	}

	// update our long version of encoder value with the 16-bits in the
	// PDO, wrapping around where needed
	unwound_cnt = priv->unwound_cnt;
	curr_cnt = le16toh(txpdo->counter_val);
	unwound_cnt += curr_cnt - priv->last_cnt;
	
	
	if ((curr_cnt & 32768) ^ (priv->last_cnt & 32768)) {
		// highest bit flipped.. see if we wrapped
		// around zero
		if ((priv->last_cnt < 16384) && (curr_cnt >= 32768 + 16384))
			unwound_cnt -= 65536;
		
		if ((priv->last_cnt >= 32768 + 16384) && (curr_cnt < 16384))
			unwound_cnt += 65536;
	}
	priv->last_cnt = curr_cnt;
	priv->unwound_cnt = unwound_cnt;

	//// construct RxPDO
	//
#if 1
	// 1600 ENC RxPDO-Map Control compact - length: 4 bytes
	rxpdo->enc_ctrl[0] = 0;
	rxpdo->enc_ctrl[1] = 0;
	rxpdo->enc_ctrl[2] = 0;
	rxpdo->enc_ctrl[3] = 0;
#endif
	
	// 1602 STM RxPDO-Map Control - length: 2 bytes
#define FLAG_STM_OUTPUT_ENABLE		(1 << 0)
#define FLAG_STM_RESET			(1 << 1)
#define FLAG_STM_REDUCE_TORQUE		(1 << 2)

	rxpdo->stm_ctrl[0] = (0 * FLAG_STM_RESET) |
	    (0 * FLAG_STM_REDUCE_TORQUE);
	
	if (INPUT(enable) == true)
		flags |= FLAG_STM_OUTPUT_ENABLE;

	if (INPUT(calib) == true) {
		// enable lower current mode during calibration
		flags |= FLAG_STM_REDUCE_TORQUE;
		// reset the current position to '0'
		priv->calib_pos = unwound_cnt;
	}

	rxpdo->stm_ctrl[0] = flags;
	rxpdo->stm_ctrl[1] = 0;

	// 1604 STM RxPDO-Map Velocity - length: 2 bytes
#if 1
	rxpdo->stm_velocity = INPUT(speed);
#else
	if (z++ & 128)
		rxpdo->stm_velocity = 130; //INPUT(speed);
	else
		rxpdo->stm_velocity = -130; //INPUT(speed);
#endif

	// jankende tandwielen motor = speed 4500 .. 6000

	priv->pos_out = unwound_cnt - priv->calib_pos;
}


static void callback_tx(struct controller_block *block, void *data)
{
	block->private->tx_buffer = data;
}

static void callback_rx(struct controller_block *block, void *data)
{
	block->private->rx_buffer = data;
}

static struct controller_block_outterm_list outterms[] = {
	{ "state_ok", CONTROLLER_BLOCK_TERM_BOOL,   offsetof(struct controller_block_private, state_ok) },
	{ "position", CONTROLLER_BLOCK_TERM_SINT32, offsetof(struct controller_block_private, pos_out)  },
	{ NULL },
};

static struct controller_block_interm_list interms[] = {
	{ "enable",    CONTROLLER_BLOCK_TERM_BOOL,   offsetof(struct controller_block_private, enable) },
	{ "calibrate", CONTROLLER_BLOCK_TERM_BOOL,   offsetof(struct controller_block_private, calib)  },
	{ "speed",     CONTROLLER_BLOCK_TERM_SINT16, offsetof(struct controller_block_private, speed)  },
	{ NULL },
};


static int pre_operational_hook(struct esc_device *esc)
{
	struct controller_block *block = esc->block;
	struct canopen_dev *can_dev;
	int ret;

	if (!block->private->candev)
		block->private->candev = esc_coe_create(esc);

	can_dev = block->private->candev;

if (1) {
	uint16_t		volt, curr, full_steps, accel, decel;
	
	ret = canopen_read_param(can_dev, REG_NOMINAL_VOLTAGE, &volt, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to read REG_NOMINAL_VOLTAGE");

	printf("nominal voltage: %dmV\n", le16toh(volt));
	
	volt = MOTOR_VOLTAGE;
	ret = canopen_write_param(can_dev, REG_NOMINAL_VOLTAGE, &volt, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_NOMINAL_VOLTAGE");

	ret = canopen_read_param(can_dev, REG_NOMINAL_VOLTAGE, &volt, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to read REG_NOMINAL_VOLTAGE");
	
	printf("nominal voltage: %dmV\n", le16toh(volt));

	curr = htole16(MOTOR_MAX_CURRENT);
	ret = canopen_write_param(can_dev, REG_MAX_CURRENT, &curr, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_MAX_CURRENT");
	
	curr = htole16(MOTOR_REDUCED_CURRENT);
	curr = htole16(MOTOR_MAX_CURRENT);
	ret = canopen_write_param(can_dev, REG_REDUCED_CURRENT, &curr, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_REDUCED_CURRENT");
	
	printf("max current: %dmA\n", le16toh(curr));

if (0) {
	uint32_t		p;
	
	p = htole32(0);
	ret = canopen_write_param(can_dev, REG_CALIB_POS, &p, 4);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_CALIB_POS");
	
	printf("max current: %dmA\n", le16toh(curr));
}

	ret = canopen_read_param(can_dev, REG_MAX_CURRENT, &curr, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to read REG_MAX_CURRENT");
	
	printf("max current: %dmA\n", le16toh(curr));

	ret = canopen_read_param(can_dev, REG_MOT_FULL_STEPS, &full_steps, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to read REG_MOT_FULL_STEPS");
	
	printf("motor full steps: %d steps/revolution\n",
	    le16toh(full_steps));


	// Acceleration / deceleration
	accel = htole16(1);
	ret = canopen_write_param(can_dev, REG_ACCEL_POS, &accel, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_ACCEL_POS");

	ret = canopen_write_param(can_dev, REG_ACCEL_NEG, &accel, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_ACCEL_NEG");

	decel = htole16(1);
	ret = canopen_write_param(can_dev, REG_DECEL_POS, &decel, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_DECEL_POS");

	ret = canopen_write_param(can_dev, REG_DECEL_NEG, &decel, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_DECEL_NEG");
	
	printf("max pos/neg accel=%d decel=%d\n", accel, decel);

}

#if 0

if (1) {
	uint16_t		stat;
	int			cnt;
	uint8_t			booltje, opmode;
	
	ret = canopen_read_param(can_dev, REG_STATUS_WORD, &stat, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to read REG_STATUS_WORD");

	printf("status word: %x\n", stat);

	for (cnt = 1; cnt < 11; cnt++) {
		ret = canopen_read_param(can_dev, 0xa010, cnt, &booltje, 1);
		if (ret != 1)
			log_send(LOG_T_ERROR, "failed to read 0xa010:%d", cnt);
		
		printf("a010:%x = %x\n", cnt, booltje);
	}

	// get actual op-mode
	ret = canopen_read_param(can_dev, 0xa010, 0x11, &opmode, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to read 0xa010:0x11", cnt);
	
	printf("actual opmode: %x\n", opmode);
}

if (1) {
	uint32_t		pos;
	
	ret = canopen_read_param(can_dev, REG_ACTUAL_POS, &pos, 4);
	if (ret != 4)
		log_send(LOG_T_ERROR, "failed to read REG_ACTUAL_POS");
	
	printf("actual pos: %d\n", pos);
}

if (1) {
	uint16_t		arg = htole16(0x8000);
	
#define REG_STM_REQUEST		0xfb00, 1

	ret = canopen_write_param(can_dev, REG_STM_REQUEST, &arg, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write REG_STM_REQUEST");
}

if (1) {
	uint8_t			reset;

	reset = 0;
	ret = canopen_read_param(can_dev, REG_OUTP_RESET, &reset, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to read REG_OUTP_RESET: %d", ret);

	reset = 0;
	ret = canopen_read_param(can_dev, REG_OUTP_ENABLE, &reset, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to read REG_OUTP_RESET: %d", ret);
}


if (1) {
	uint16_t		stat;
	int			cnt;
	uint8_t			booltje;
	
	ret = canopen_read_param(can_dev, REG_STATUS_WORD, &stat, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to read REG_STATUS_WORD");

	for (cnt = 1; cnt < 11; cnt++) {
		ret = canopen_read_param(can_dev, 0xa010, cnt, &booltje, 1);
		if (ret != 1)
			log_send(LOG_T_ERROR, "failed to read 0xa010:%d", cnt);
		
		printf("a010:%x = %x\n", cnt, booltje);
	}
	
	printf("status word: %x\n", stat);
}

if (1) {
	int			cnt;
	uint8_t			val;
	
	for (cnt = 1; cnt <= 7; cnt++) {
		ret = canopen_read_param(can_dev, 0x6010, cnt, &val, 1);
		if (ret != 1)
			log_send(LOG_T_ERROR, "failed to read 0x6010:%d", cnt);
		
		printf("6010:%x = %x\n", cnt, val);
	}
	for (cnt = 0xc; cnt <= 0xe; cnt++) {
		ret = canopen_read_param(can_dev, 0x6010, cnt, &val, 1);
		if (ret != 1)
			log_send(LOG_T_ERROR, "failed to read 0x6010:%d", cnt);
		
		printf("6010:%x = %x\n", cnt, val);
	}
}

if (1) {
	int			cnt;
	uint8_t			val;
	
	for (cnt = 1; cnt <= 7; cnt++) {
		ret = canopen_read_param(can_dev, 0x6020, cnt, &val, 1);
		if (ret != 1)
			log_send(LOG_T_ERROR, "failed to read 0x6020:%d", cnt);
		
		printf("6020:%x = %x\n", cnt, val);
	}
	for (cnt = 0x11; cnt <= 0x11; cnt++) {
		ret = canopen_read_param(can_dev, 0x6020, cnt, &val, 1);
		if (ret != 1)
			log_send(LOG_T_ERROR, "failed to read 0x6020:%d", cnt);
		
		printf("6020:%x = %x\n", cnt, val);
	}
	for (cnt = 0x21; cnt <= 0x22; cnt++) {
		ret = canopen_read_param(can_dev, 0x6020, cnt, &val, 1);
		if (ret != 1)
			log_send(LOG_T_ERROR, "failed to read 0x6020:%d", cnt);
		
		printf("6020:%x = %x\n", cnt, val);
	}
}

#endif

	// by default the RxPDO (PLC -> motor controller) is assigned to:
	// 0x1600, 0x1602, 0x1604
	// this can be changed by writing to reg 0x1c12:1..3 (RxPDO assign)

if (1) {
	uint8_t			num_assigned;
	uint16_t		pdo_index;
	
	num_assigned = 0;
	ret = canopen_write_param(can_dev, 0x1c12, 0, &num_assigned, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 0");

	// assign reg 1600 (STM RxPDO-Map Control) as first part of the PDO
	pdo_index = htole16(0x1600);
	ret = canopen_write_param(can_dev, 0x1c12, 1, &pdo_index, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 1");

	// assign reg 1602 (STM RxPDO-Map Control) as first part of the PDO
	pdo_index = htole16(0x1602);
	ret = canopen_write_param(can_dev, 0x1c12, 2, &pdo_index, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 1");

	// assign reg 1604 (STM RxPDO-Map Velocity) as second part of the PDO
	pdo_index = htole16(0x1604);
	ret = canopen_write_param(can_dev, 0x1c12, 3, &pdo_index, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 2");

	// .. and enable just this part of the PDO
	num_assigned = 3;
	ret = canopen_write_param(can_dev, 0x1c12, 0, &num_assigned, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 0 to 1");

}

if (0) {
	// try to request RxPDO 1602, 1603
	uint8_t			num_assigned;
	uint16_t		pdo_index;
	
	num_assigned = 0;
	ret = canopen_write_param(can_dev, 0x1c12, 0, &num_assigned, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 0");

	// assign reg 1602 (STM RxPDO-Map Control) as first part of the PDO
	pdo_index = htole16(0x1602);
	ret = canopen_write_param(can_dev, 0x1c12, 1, &pdo_index, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 1");

	// assign reg 1603 (STM RxPDO-Map Control) as second part of the PDO
	pdo_index = htole16(0x1603);
	ret = canopen_write_param(can_dev, 0x1c12, 2, &pdo_index, 2);
	if (ret != 2)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 2");

	// .. and enable this part of the PDO
	num_assigned = 1;
	ret = canopen_write_param(can_dev, 0x1c12, 0, &num_assigned, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 0 to 2");

}

	// show RxPDO assignments (motor controller -> PLC)
if (1) {
	uint8_t			num_assigned = 0;
	uint16_t		pdo_index;
	int			cnt;
	
	ret = canopen_read_param(can_dev, 0x1c12, 0, &num_assigned, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to write RxPDO index reg 0");
	printf("RxPDO's assigned: %d\n", num_assigned);
	
	for (cnt = 1; cnt <=3; cnt++) {
		ret = canopen_read_param(can_dev, 0x1c12, cnt, &pdo_index, 2);
		if (ret != 2)
			log_send(LOG_T_ERROR, "failed to read RxPDO index reg %d", cnt);
		
		printf("RxPDO [%d] index=%x\n", cnt, pdo_index);
	}
}
	usleep(200 * 1000);


	// show TxPDO assignments (motor controller -> PLC)
if (1) {
	uint8_t			num_assigned = 0;
	uint16_t		pdo_index;
	int			cnt;
	
	ret = canopen_read_param(can_dev, 0x1c13, 0, &num_assigned, 1);
	if (ret != 1)
		log_send(LOG_T_ERROR, "failed to write TxPDO index reg 0");
	printf("TxPDO's assigned: %d\n", num_assigned);
	
	for (cnt = 1; cnt <=2; cnt++) {
		ret = canopen_read_param(can_dev, 0x1c13, cnt, &pdo_index, 2);
		if (ret != 2)
			log_send(LOG_T_ERROR, "failed to write TxPDO index reg %d", cnt);
		
		printf("TxPDO [%d] index=%x\n", cnt, pdo_index);
	}
}

	return 0;
}



struct controller_block * block_beckhoff_el7031_create(char *name, va_list ap)
{
	struct controller_block		*block;
	int				devno;
	char				*busname;
	struct esc_device		*esc;
	uint32_t vendor, product;

	busname = va_arg(ap, char *);
	devno = va_arg(ap, int);
	
	esc = esc_device_create(busname, devno, name);
	
	vendor = esc_esi_vendorid_get(esc);
	product = esc_esi_productcode_get(esc);
		
	printf("scan: vendor=%x, product=%x, devno=%d\n",
	    vendor, product, devno);
		
	if (vendor != ESC_ESI_VENDORID_BECKHOFF ||
	    product != ESC_ESI_PRODUCTCODE_BECKHOFF_EL7031) {
		log_send(LOG_T_ERROR, "%s: el7031: Device not found", name);
		return NULL;
	}
	
	block = controller_block_alloc("beckhoff_el7031", name,
	    sizeof(struct controller_block_private));
	if (block == NULL)
		return NULL;

	block->private->esc = esc;
	esc_esi_device_fill(esc);
	esc->block = block;
	esc->block_rx = block;
	esc->block_tx = block;
	esc->client_tx->callback = callback_tx;
	esc->client_rx->callback = callback_rx;
	esc->pre_operational_hook = pre_operational_hook;
	esc_device_initialize_operational(esc);
	

	if (controller_block_outterm_list_init(block, outterms))
		goto err_outputs;	
	if (controller_block_interm_list_init(block, interms))
		goto err_inputs;	

	block->private->state_ok = false;
	block->private->unwound_cnt = 0;
	block->private->last_cnt = 0;
	block->private->calib_pos = 0;
	
	block->calculate = calculate;
	block->param_get = NULL;
	block->param_set = NULL;

	controller_block_add(block);
	return block;

err_outputs:
err_inputs:
	controller_block_free(block);
	return NULL;
}
