/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2012, 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2012, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <string.h>
#include <stdio.h>

#include <controller/controller_block.h>
#include <ec/esc.h>
#include <log/log.h>
#include <ec/esc_esi.h>
#include <ec/canopen.h>
#include <ec/esc_device.h>
#include <ec/esc_coe.h>
#include <ec/esc_id.h>



#define STOEBER_PARAM2INDEX(group, nr) (0x2000 + ((group) - 'A') * 0x200 + nr)

#define STOEBER_PDO_MAP(g,nr,s,l) \
    htole32((STOEBER_PARAM2INDEX(g,nr)<<16)|((s)<<8)|l)

enum stoeber_e48 {
	STOEBER_E48_SELFTEST = 		0,
	STOEBER_E48_SWITCHON_DISABLE =	1,
	STOEBER_E48_READY_SWITCHON =	2,
	STOEBER_E48_SWTICHED_ON =	3,
	STOEBER_E48_ENABLED =		4,
	STOEBER_E48_FAULT =		5,
	STOEBER_E48_FAULT_REACTION =	6,
	STOEBER_E48_QUICKSTOP =		7,
};

static char *stoeber_e48_tostring(enum stoeber_e48 e)
{
	switch(e) {
		case STOEBER_E48_SELFTEST:
			return "Self-test";
		case STOEBER_E48_SWITCHON_DISABLE:
			return "Switch-on disable";
		case STOEBER_E48_READY_SWITCHON:
			return "Ready for switch-on";
		case STOEBER_E48_SWTICHED_ON:
			return "Switched on";
		case STOEBER_E48_ENABLED:
			return "Enabled";
		case STOEBER_E48_FAULT:
			return "Fault";
		case STOEBER_E48_FAULT_REACTION:
			return "Fault reaction";
		case STOEBER_E48_QUICKSTOP:
			return "Quick stop";
	}
}


#define RPM2RADS(val) ((val)*2.0*M_PI/60.0)

#define static_assert(e) \
do { \
	enum { assert_static__ = 1/(e) }; \
} while (0)

/* rx pdo */
struct rx_pdo {
	uint32_t I80; /* current position */
	uint32_t E91; /* motor speed */
	uint16_t E90; /* motor torque */
	uint16_t E19; /* binary inputs */
	uint16_t E10; /* AE1 level */
	uint8_t E48; /* device control state */
} __packed;
	
/* tx pdo */
struct tx_pdo {
	uint8_t A180; /* device control byte */
	uint8_t F210; /* BA control bits */
	uint16_t I210; /* posi.control word */
	uint32_t I213; /* target position */
	uint16_t I215; /* speed */
	uint16_t C230; /* torque limit */
} __packed;

struct controller_block_private {
	float max_speed;
	float max_accel;
	float max_torque;

	float standstill_torque;
	float brake_resistor;
	float brake_power;
	uint32_t serial_number;
	
	bool enabled;
	enum stoeber_e48 control_state;
	
	float output_position;
	float output_speed;
	float output_torque;
	bool output_be1;
	bool output_be2;
	bool output_be3;
	bool output_be4;
	bool output_be5;
	bool output_external_enable;
	float output_ae1;
	
	float *input_speed;
	float *input_torque;
	bool *input_enable;
	bool *input_ba1;
	bool *input_ba2;

	struct tx_pdo *tx_pdo;
	struct rx_pdo *rx_pdo;

	struct esc_device *dev;
	struct canopen_dev *candev;
};

static void calculate_tx(struct controller_block *tx)
{
	struct controller_block_private *private = tx->private;
	bool enabled;
	float sp;
	float torque;

	enabled = *private->input_enable;
	enabled &= private->output_external_enable;
	
	private->tx_pdo->A180 = enabled;
	private->tx_pdo->F210 = (*private->input_ba1 << 0) | (*private->input_ba2 << 1);
	private->tx_pdo->I210 = htole16(le16toh(private->tx_pdo->I210) ^ 1);
	
	if (enabled)
		sp = *private->input_speed * 16384.0 / RPM2RADS(3000.0);
	else
		sp = 0;
	
	if (sp > 32767)
		sp = 32767;
	if (sp < -32768)
		sp = -32768;
	
	private->tx_pdo->I215 = htole16((int16_t)sp);
	
	torque = *private->input_torque / private->standstill_torque * 16384;

	private->tx_pdo->C230 = htole16((int16_t)torque);
}

static void calculate_rx(struct controller_block *rx)
{
	struct controller_block_private *private = rx->private;
	struct rx_pdo *rx_pdo;
	int ret;
	float speed, torque, position, ae1;
	bool enabled, be1, be2, be3, be4, be5;
	bool external_enable;
	uint16_t E19;
	enum stoeber_e48 e48;
	
	enabled = private->enabled;
	
	if (ret == 0) {
		rx_pdo = private->rx_pdo;
		
		position = (int32_t)le32toh(rx_pdo->I80);
		position *= 2.0 * M_PI / 16384;
		
		speed = (int32_t)le32toh(rx_pdo->E91);
		speed *= 1.0/16384;
		speed = RPM2RADS(speed);
		
		torque = (int16_t)le16toh(rx_pdo->E90);
		torque *= 1.0 / 1200;
		
		ae1 = (int16_t)le16toh(rx_pdo->E10);
		ae1 *= 20.0 / 32767.0;
		
		E19 = le16toh(rx_pdo->E19);
		be1 = E19 & 0x01;
		be2 = E19 & 0x02;
		be3 = E19 & 0x04;
		be4 = E19 & 0x08;
		be5 = E19 & 0x10;
		
		e48 = rx_pdo->E48;
		if (e48 != private->control_state) {
			enum log_type lt;
			
			switch(e48) {
				case STOEBER_E48_SELFTEST:
				case STOEBER_E48_SWITCHON_DISABLE:
				case STOEBER_E48_READY_SWITCHON:
				case STOEBER_E48_SWTICHED_ON:
					if (private->control_state ==
					    STOEBER_E48_ENABLED)
						lt = LOG_T_ERROR;
					else
						lt = LOG_T_WARNING;
					break;
				case STOEBER_E48_ENABLED:
					lt = LOG_T_WARNING;
					break;
				case STOEBER_E48_FAULT:
				case STOEBER_E48_FAULT_REACTION:
				case STOEBER_E48_QUICKSTOP:
				default:
					lt = LOG_T_ERROR;
					break;
			}
			
			log_send(lt, "%s: State changed from '%s' to '%s'",
			   rx->name,
			   stoeber_e48_tostring(private->control_state),
			   stoeber_e48_tostring(e48));
			private->control_state = e48;
		}
		
		external_enable = e48 & STOEBER_E48_ENABLED;
		if (external_enable != private->output_external_enable) {
			log_send(LOG_T_WARNING, 
			    "%s: external enable changed state to %s",
			    rx->name, external_enable ? "true" : "false");
		}
		if (!external_enable & enabled) {
			log_send(LOG_T_ERROR,
			    "%s: external_enable prevents device from keeping enabled state",
			    rx->name);
			enabled = false;
		}
	} else {
		enabled = false;
		speed = 0;
		torque = 0;
		position = 0;
		be1 = false;
		be2 = false;
		be3 = false;
		be4 = false;
		be5 = false;
		ae1 = 0;
		external_enable = false;
	}
	
	private->output_position = position;
	private->output_speed = speed;
	private->output_torque = torque;
	private->enabled = enabled;
	private->output_be1 = be1;
	private->output_be2 = be2;
	private->output_be3 = be3;
	private->output_be4 = be4;
	private->output_be5 = be5;
	private->output_ae1 = ae1;
	private->output_external_enable = external_enable;
}

static void callback_tx(struct controller_block *block, void *data)
{
	block->private->tx_pdo = data;
}

static void callback_rx(struct controller_block *block, void *data)
{
	block->private->rx_pdo = data;
}


static struct controller_block_interm_list interms[] = {
	{ "speed",  CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input_speed)  },
	{ "torque", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, input_torque) },
	{ "enable", CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, input_enable) },
	{ "ba1",    CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, input_ba1)    },
	{ "ba2",    CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, input_ba2)    },
	{ NULL },
};

static struct controller_block_outterm_list outterms[] = {
	{ "position", CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, output_position) },
	{ "speed",    CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, output_speed)    },
	{ "torque",   CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, output_torque)   },
	{ "enabled",  CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, enabled)         },
	{ "be1",      CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, output_be1)      },
	{ "be2",      CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, output_be2)      },
	{ "be3",      CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, output_be3)      },
	{ "be4",      CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, output_be4)      },
	{ "be5",      CONTROLLER_BLOCK_TERM_BOOL,  offsetof(struct controller_block_private, output_be5)      },
	{ "ae1",      CONTROLLER_BLOCK_TERM_FLOAT, offsetof(struct controller_block_private, output_ae1)      },
	{ "external_enable", CONTROLLER_BLOCK_TERM_BOOL, offsetof(struct controller_block_private, output_external_enable) },
	{ NULL },
};

static void stoeber_emergency_handler(struct canopen_dev *dev, 
    uint16_t code, uint8_t reg, uint8_t vendor[5])
{
	char *msg = "unknown";
	
	if (code == 0x0000) {
		msg = "no_error";
	} else if (code == 0x2110) {
		msg = "short circuit earth";
	} else if (code == 0x2230) {
		msg = "intern short circuit earth";
	} else if (code == 0x2310) {
		msg = "continuous overcurrent";
	} else if (code == 0x3110) {
		msg = "mains overvoltage";
	} else if (code == 0x3120) {
		msg = "mains undervoltage";
	} else if (code == 0x4210) {
		msg = "temperature device";
	} else if (code == 0x4280) {
		msg = "temperature device i2t";
	} else if (code == 0x4310) {
		msg = "temperature drive";
	} else if (code == 0x4310) {
		msg = "temperature drive i2t";
	} else if (code == 0x5000) {
		msg = "device hardware";
	} else if (code == 0x5200) {
		msg = "device hardware control";
	} else if (code == 0x6010) {
		msg = "software reset";
	} else if (code == 0x6100) {
		msg = "internal software";
	} else if (code == 0x6200) {
		msg = "user software";
	} else if (code == 0x6300) {
		msg = "data record";
	} else if (code == 0x6310) {
		msg = "loss of parameters";
	} else if (code == 0x7110) {
		msg = "brake chopper";
	} else if (code == 0x7120) {
		msg = "motor";
	} else if (code == 0x7303) {
		msg = "resolver 1 fault";
	} else if (code == 0x8100) {
		msg = "communication";
	} else if (code == 0x8400) {
		msg = "velocity speed control";
	} else if (code == 0x8311) {
		msg = "excess torque";
	} else if (code == 0x9000) {
		msg = "external error";
	} else if (code == 0xa000) {
		msg = "Transition from Pre-Operational to Safe-Operational was not successful";
	} else if (code == 0xa001) {
		msg = "Transition from Safe-Operational to Pre-Operational was not successful";
	}
	
	log_send(LOG_T_WARNING, "%s: %s",
	    dev->bus->name, msg);
}

static int pre_operational_hook(struct esc_device *esc)
{
	struct controller_block *block = esc->block;
	struct canopen_dev *canopen_dev;
	uint8_t val8;
	uint16_t leval16;
	uint32_t leval32;
	
	if (!block->private->candev) {
		block->private->candev = esc_coe_create(esc);

		canopen_emergency_handler_attach(block->private->candev, stoeber_emergency_handler);
	}
	canopen_dev = block->private->candev;

	/* A180 device control, disable it for now */
	val8 = 0x00;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 180), 0x0, &val8, 1);

	/* A61 fault reset source to parameter */
	val8 = 2;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 61), 0x0, &val8, 1);
	usleep(3000);
	/* A180 device control, reset faults */
	val8 = 0x02;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 180), 0x0, &val8, 1);
	usleep(50000);
	/* A180 device control, reset faults */
	val8 = 0x00;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 180), 0x0, &val8, 1);

	canopen_read_param(canopen_dev, STOEBER_PARAM2INDEX('E', 52), 0x0, &leval32, 4);
	block->private->serial_number = le32toh(leval32);

	/* A21 brake resistor */
	leval16 = htole16(block->private->brake_resistor * 10.0);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 21), 0x0, &leval16, 2);
	/* A22 brake resistor power */
	leval16 = htole16(block->private->brake_power);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 22), 0x0, &leval16, 2);

	/* A29 fault quickstop */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 29), 0x0, &val8, 1);

	/* A34 auto start */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 34), 0x0, &val8, 1);
	
	/* A45*/
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 45), 0x0, &val8, 1);
	
	/* A60*/
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 60), 0x0, &val8, 1);
	
	/* A150 cycle time */
	val8 = 0x06;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 150), 0x0, &val8, 1);

	/* A258 set PDO timeout */
	leval16 = htole16(1000);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 258), 0x0, &leval16, 2);

	/* G90 activate PLL */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 90), 0x0, &val8, 1);

	/* G91 pll phase offset */
	leval32 = htole32(0);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 91), 0x0, &leval32, 4);

	/* G92 pll gain pk */
	leval32 = htole32(10);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 92), 0x0, &leval32, 4);

	/* G93 pll low pass */
	leval32 = htole32(999);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 93), 0x0, &leval32, 4);

	/* G98 cycle to 4ms */
	leval16 = htole16(4000);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('G', 98), 0x0, &leval16, 2);

	/* Set I06 to 2 decimal digits */
	val8 = 2;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 6), 0x0, &val8, 1);

	/* C32 (motor integrator) to 3000 ms */
	leval16 = htole16(32000);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 32), 0x0, &leval16, 2);

	/* C33 low pass reference speed 1 ms */
	leval32 = htole32(10);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 33), 0x0, &leval32, 4);

	/* C130 torque limit source to parameter */
	val8 = 4;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 130), 0x0, &val8, 1);

	/* Set I10 to maximum speed */
	leval32 = htole32(block->private->max_speed / (2*M_PI) * 360 * 100);
//	leval32 = htole32(3000 * 360 / 60 * 100);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 10), 0x0, &leval32, 4);

	/* Set I11 to maximum acceleration */
//	leval32 = htole32(2000000 * 100);
	leval32 = htole32(block->private->max_accel * 100);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 11), 0x0, &leval32, 4);
		
	/* Set I25 to 100% speed ff */
	leval16 = htole16(100);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 25), 0x0, &leval16, 2);

	/* Set I30 to 1 (encoder signal 0) */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 30), 0x0, &val8, 1);

	/* Set I37 to 1 (active) */
	val8 = 1;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 37), 0x0, &val8, 1);

	/* Set I100 to parameter (execute source) */
	val8 = 2;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 100), 0x0, &val8, 1);
		
	/* B17 Torque M0 */
	canopen_read_param(canopen_dev, STOEBER_PARAM2INDEX('B', 17), 0x0, &leval32, 4);
	log_send(LOG_T_DEBUG, "B17: %08x %d", le32toh(leval32), le32toh(leval32));
	block->private->standstill_torque = (float)(le32toh(leval32)) / 1000.0;
	log_send(LOG_T_DEBUG, "Standstill torque: %f", block->private->standstill_torque);

	/* C03 max pos torque 150% of standstill */
	leval32 = htole32(150);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 3), 0x0, &leval32, 4);

	/* C05 max neg torque -150% of standstill */
	leval32 = htole32(-150);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('C', 5), 0x0, &leval32, 4);

//	/* A180 device control, enable it to make sure we stay
//	   at the same position */
//	val8 = 0x01;
//	canopen_write_param(canopen_dev, 0x20b4, 0x0, &val8, 1);

	/* I211 command byte to 4 (move_velocity) */
	val8 = 0x04;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 211), 0x0, &val8, 1);
	/* I215 v to 0 */
	leval16 = htole16(0);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 215), 0x0, &leval16, 2);
	/* I210 latch command */
	val8 = 0x00;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 210), 0x0, &val8, 1);
	val8 = 0x01;
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('I', 210), 0x0, &val8, 1);

	/* Transmit pdo: */

	/* I80 (4bytes) */
	leval32 = STOEBER_PDO_MAP('I', 80, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x0, &leval32, 4);
		
	/* E91 (4bytes) */
	leval32 = STOEBER_PDO_MAP('E', 91, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x1, &leval32, 4);

	/* E90 (2bytes) */
	leval32 = STOEBER_PDO_MAP('E', 90, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x2, &leval32, 4);
		
	/* E19 (2bytes) */
	leval32 = STOEBER_PDO_MAP('E', 19, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x3, &leval32, 4);

	/* E10 (2bytes) */
	leval32 = STOEBER_PDO_MAP('E', 10, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x4, &leval32, 4);

	/* E48 (1byte) */
	leval32 = STOEBER_PDO_MAP('E', 48, 0, 8);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 233), 0x5, &leval32, 4);

	/* Receive pdo: */

	/* A180 (1byte) */
	leval32 = STOEBER_PDO_MAP('A', 180, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x0, &leval32, 4);
		
	/* F210 (1byte) */
	leval32 = STOEBER_PDO_MAP('F', 210, 0, 32);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x1, &leval32, 4);

	/* I210 (2bytes) */
	leval32 = STOEBER_PDO_MAP('I', 210, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x2, &leval32, 4);
		
	/* I213 (4bytes) */
	leval32 = STOEBER_PDO_MAP('I', 213, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x3, &leval32, 4);

	/* I215 (2bytes) */
	leval32 = STOEBER_PDO_MAP('I', 215, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x4, &leval32, 4);

	/* C230 (2bytes) */
	leval32 = STOEBER_PDO_MAP('C', 230, 0, 16);
	canopen_write_param(canopen_dev, STOEBER_PARAM2INDEX('A', 225), 0x5, &leval32, 4);

	/* 'wishfull thinking' -> the sample function will check against this later */
	block->private->control_state = STOEBER_E48_ENABLED;

	return 0;
}

struct controller_block *block_stoeber_create(char *name, va_list ap)
{
	struct controller_block *stoeber;
	struct controller_block *stoeber_tx;
	struct controller_block *stoeber_rx;
	struct controller_block_private *private;
	int ec_pos;
	char *busname;
	int i;
	int serial_number;
	double max_torque, brake_resistor, brake_power;
	char trx_name[strlen(name)+strlen("_Xx0")];
	struct esc_device *esc;
	uint32_t vendorid, productcode, serialno;
	
	busname = va_arg(ap, char *);
	ec_pos = va_arg(ap, int);
	serial_number = va_arg(ap, int);
	max_torque = va_arg(ap, double);
	brake_resistor = va_arg(ap, double);
	brake_power = va_arg(ap, double);

	stoeber = controller_block_alloc("stoeber", name, 
	    sizeof(struct controller_block_private));
	if (!stoeber)
		goto err_malloc;
	
	sprintf(trx_name, "%s_tx", name);
	stoeber_tx = controller_block_alloc("stoeber_tx", name, 0);
		goto err_malloc_tx;
	
	stoeber_tx->private = stoeber->private;
	
	sprintf(trx_name, "%s_rx", name);
	stoeber_rx = controller_block_alloc("stoeber_rx", name, 0);
	if (!stoeber_rx)
		goto err_malloc_rx;
	
	stoeber_rx->private = stoeber->private;
	

	esc = esc_device_create(busname, ec_pos, name);

	stoeber->private->dev = esc;

	vendorid = esc_esi_vendorid_get(esc);
	productcode = esc_esi_productcode_get(esc);
	serialno = esc_esi_serialno_get(esc);

	log_send(LOG_T_DEBUG, 
	    "VendorId: 0x%04x, ProductCode: 0x%04x, SerialNo: 0x%04x",
	    vendorid, productcode, serialno);

	if (vendorid != ESC_ESI_VENDORID_STOEBER ||
	    productcode != ESC_ESI_PRODUCTCODE_STOEBER_XDS5000) {
		log_send(LOG_T_ERROR, "%s: Device not found\n", name);
		return NULL;
	}


	private->max_speed = RPM2RADS(3000);
	private->max_accel = 3000 * 360 / 60 * 10;
	private->max_torque = max_torque;
	
	private->brake_resistor = brake_resistor;
	private->brake_power = brake_power;

	esc_esi_device_fill(esc);
	esc->block = stoeber;
	esc->block_tx = stoeber_tx;
	esc->block_rx = stoeber_rx;
	esc->callback_tx = callback_tx;
	esc->callback_rx = callback_rx;
	esc->pre_operational_hook = pre_operational_hook;
	
	/* sm2 tx pdo (master->slave) @0x1600 */
	esc->tx_pdo_watchdog = false;
	esc->tx_pdo = true;
	esc->tx_pdo_sm = 2;
	esc->sm[2].start = 0x1600;
	esc->sm[2].len = sizeof(struct tx_pdo);
	esc->sm[2].enabled = true;
	static_assert(0xc == sizeof(struct tx_pdo));

	/* sm3 rx pdo (slave->master) @0x1a00 */
	esc->rx_pdo = true;
	esc->rx_pdo_sm = 3;
	esc->sm[3].start = 0x1a00;
	esc->sm[3].len = sizeof(struct rx_pdo);
	esc->sm[3].enabled = true;
	static_assert(0xf == sizeof(struct rx_pdo));

	/* mailbox read (slave->master) @0x18f6 */
	esc->sm[1].start = 0x18f6;
	esc->sm[1].len = 0x0040;
	esc->mbox_rd_sm = 1;
	
	/* mailbox write (master->slave) @0x1800 */
	esc->sm[0].start = 0x1800;
	esc->sm[0].len = 0x0040;
	esc->mbox_wr_sm = 0;

	esc->dc_sync.sync0_cycle = 4000 * 1000;
	esc->dc_sync.sync0_shift = 0;
	esc->dc_sync.sync_out_active = true;
	esc->dc_sync.sync0_active = true;
	esc->dc_sync.sync1_active = false;


	esc_device_initialize_operational(esc);
	

	log_send(LOG_T_DEBUG, "Drive serial number: %d",
	    private->serial_number);
	if (serial_number && serial_number != private->serial_number) {
		log_send(LOG_T_ERROR, 
		    "Serial number of drive (%d) does not match requested number (%d)",
		    private->serial_number, serial_number);
		goto err_stoeber_init;
	}	

	if (controller_block_outterm_list_init(stoeber_rx, outterms))
		goto err_rx_output;

	/* stoeber also exports these outputs, we will re-use the same struct.
	   This is possible as both source and sourceterm point to the
	   stoeber_tx structs already. */
	stoeber->outputs = stoeber_rx->outputs;
	stoeber->output = stoeber_rx->output;

	if (controller_block_interm_list_init(stoeber_tx, interms))
		goto err_tx_input;

	stoeber->inputs = stoeber_tx->inputs;
	stoeber->input = calloc(stoeber->inputs, sizeof(struct controller_block_interm));
	if (!stoeber->input)
		goto err_input;
	
	for (i = 0; i < stoeber->inputs; i++) {
		stoeber->input[i].name = stoeber_tx->input[i].name;
		stoeber->input[i].type = stoeber_tx->input[i].type;
		stoeber->input[i].value = stoeber_tx->input[i].value;
		stoeber->input[i].ghostof = &stoeber_tx->input[i];
	}

	stoeber_tx->calculate = calculate_tx;
	stoeber_rx->calculate = calculate_rx;
	
	controller_block_add(stoeber_tx);
	controller_block_add(stoeber_rx);
	controller_block_add(stoeber);

	return stoeber;

err_input:
	free(stoeber_tx->input);
err_tx_input:
	free(stoeber->output);
err_rx_output:
	/* no way to un-init the device? */
err_stoeber_init:
	free(stoeber_rx->name);
	free(stoeber_rx);
err_malloc_rx:
	free(stoeber_tx->name);
	free(stoeber_tx);
err_malloc_tx:
	free(stoeber->private);
	free(stoeber->name);
	free(stoeber);
err_malloc:
	return NULL;
}
