/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2009
	Copyright Stichting C.A. Muller Radioastronomiestation, 2009   

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <string.h>
#include <stdio.h>
#include <time.h>

#include <ec/esc.h>
#include <ec/esc_registers.h>
#include <ec/ec.h>
#include <ec/canopen.h>
#include <ec/esc_device.h>
#include <log/log.h>


bool esc_timeout_expired(struct timespec *start, struct timespec *timeout)
{
	time_t sec;
	long nanosec;
	struct timespec t_now;
	
	sec = start->tv_sec + timeout->tv_sec;
	nanosec = start->tv_nsec + timeout->tv_nsec;
	
	if (nanosec > 1000000000) {
		sec++;
		nanosec -= 1000000000;
	}
	clock_gettime(CLOCK_MONOTONIC, &t_now);
	
	/* seconds expired? */
	if (t_now.tv_sec < sec)
		return false;
	if (t_now.tv_sec > sec)
		return true;

	/* within one second, check nanoseconds*/
	if (t_now.tv_nsec > nanosec)
		return true;
	
	return false;
}

enum esc_al_state esc_al_state_get(struct esc_device *esc)
{
	uint16_t state;
	struct ec_dgram_addr state_add;
	int ret;
	
	memcpy(&state_add, &esc->addr, sizeof(state_add));
	state_add.addr.position.off = ESC_ADDR_MAP_AL_STATUS;
	ret = ec_datagram_read(&state_add, &state, 2);

	if (ret == 2)
		return le16toh(state) & 0x000f;

	ret = ec_datagram_read(&state_add, &state, 2);

	if (ret == 2) {
		return le16toh(state) & 0x000f;
	} else
		return ESC_AL_STATE_ERROR;
}

int esc_al_state_set(struct esc_device *esc, enum esc_al_state newstate,
    struct timespec *timeout)
{
	uint16_t state;
	struct ec_dgram_addr state_add;
	int ret;
	struct timespec t_start;
	enum esc_al_state rd_state;
	
	memcpy(&state_add, &esc->addr, sizeof(state_add));
	state_add.addr.position.off = ESC_ADDR_MAP_AL_CONTROL;
	state = htole16(newstate);
	clock_gettime(CLOCK_MONOTONIC, &t_start);
	do {
		ret = ec_datagram_write(&state_add, &state, sizeof(uint16_t));
	} while (ret == 0 && !esc_timeout_expired(&t_start, timeout));
	
	if (ret == 2) {
		do {
			enum esc_al_status status;
			
			rd_state = esc_al_state_get(esc);
			if (rd_state == newstate)
				return newstate;
			else if (rd_state == ESC_AL_STATE_ERROR)
				log_send(LOG_T_ERROR, 
				    "%s: Could not read state (trying to set %d %s)",
				    esc->name,
				    newstate,
				    esc_al_state2str(newstate));
				    
			status = esc_al_status_code_get(esc);
			if (status < 0) {
				log_send(LOG_T_ERROR,
				    "%s: Could not read status",
				    esc->name);
				return -1;
			} else if (status != 0) {
				log_send(LOG_T_DEBUG, 
				    "%s: status code: 0x%x %s", esc->name,
				    status, esc_al_status2str(status));
				esc_al_error_ack(esc);
				
				return -1;
			}
		} while (!esc_timeout_expired(&t_start, timeout));
		log_send(LOG_T_ERROR, 
		    "%s: Timeout while trying to set state %d %s, stuck in state %d %s",
		    esc->name,
		    newstate, esc_al_state2str(newstate), 
		    rd_state, esc_al_state2str(rd_state));
		/* timeout... */
		return rd_state;
	} else
		return -1;
}

enum esc_al_status esc_al_status_code_get(struct esc_device *esc)
{
	uint16_t state;
	struct ec_dgram_addr state_add;
	int ret;
	
	memcpy(&state_add, &esc->addr, sizeof(state_add));
	state_add.addr.position.off = ESC_ADDR_MAP_AL_STATUS;
	ret = ec_datagram_read(&state_add, &state, 2);
	if (ret != 2)
		goto err_get_state;
	
	if (le16toh(state) & ESC_ADDR_MAP_AL_STATUS_ERR_IND) {
		state_add.addr.position.off = ESC_ADDR_MAP_AL_STATUS_CODE;
		ret = ec_datagram_read(&state_add, &state, 2);
		state = le16toh(state);
	} else {
		state = ESC_AL_STATUS_NO_ERROR;
	}
	
	return state;

err_get_state:
	return ESC_AL_STATUS_ERROR;
}

int esc_al_error_ack(struct esc_device *esc)
{
	uint16_t ctrl;
	struct ec_dgram_addr state_add;
	int ret;
	enum esc_al_state state;
	
	memcpy(&state_add, &esc->addr, sizeof(state_add));

	state = esc_al_state_get(esc);
	if (state == ESC_AL_STATE_ERROR)
		return -1;
	ctrl = state;
	
	if (!state)
		return 0;
		
	state_add.addr.position.off = ESC_ADDR_MAP_AL_CONTROL;
	ctrl |= htole16(ESC_ADDR_MAP_AL_CONTROL_ERROR_ACK);
	ret = ec_datagram_write(&state_add, &ctrl, 2);
	
	if (ret == 2)
		return 0;
	else
		return -1;
}

int esc_init(struct esc_device *esc)
{
	uint8_t nr_syncmgr;
	int ret, i;
	struct timespec	timeout = { 1, 0 };
	struct ec_dgram_addr addr;
	
	memcpy(&addr, &esc->addr, sizeof(struct ec_dgram_addr));

	if (esc_al_state_set(esc, ESC_AL_STATE_INIT, &timeout) < 0)
		log_send(LOG_T_ERROR, "Could not go to state init");

	addr.addr.position.off = ESC_ADDR_MAP_SYNCMANAGERS_SUPPORTED;
	ret = ec_datagram_read(&addr, &nr_syncmgr, 1);
	if (ret != 1) {
		log_send(LOG_T_ERROR, "Could not read ESC information");
		return -1;
	}

	log_send(LOG_T_DEBUG, "ESC supports %d SyncManagers", nr_syncmgr);
	/* De-activate all sync managers. */
	for (i = 0; i < nr_syncmgr; i++) {
		uint16_t off = ESC_ADDR_MAP_SYNCMANAGER_NR(i);
		uint8_t val8;
		uint16_t val16;
		
		addr.addr.position.off = off | ESC_ADDR_MAP_SYNCMANAGER_ACTIVATE;
		val8 = 0;
		
		ret = ec_datagram_write(&addr, &val8, 1);
		if (ret != 1) {
			log_send(LOG_T_ERROR, 
			    "Could not deactivate sync manager %d", i);
			return -1;
		}

		val16 = 0;
		addr.addr.position.off = off | ESC_ADDR_MAP_SYNCMANAGER_LENGTH;
			ec_datagram_write(&addr, &val16, 2);
	}

	log_send(LOG_T_DEBUG, "ACK any errors");
	/* Try to ack any errors */
	esc_al_error_ack(esc);

	return 0;
}


