/*
	Copyright Jeroen Vreeken (pe1rxq@amsat.org), 2013
	Copyright Stichting C.A. Muller Radioastronomiestation, 2013

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
/*
	esc_device.c
	
	Code for using and maintaining the esc_device.	
 */

#include <string.h>
#include <limits.h>

#include <ec/ec.h>
#include <ec/esc.h>
#include <ec/esc_device.h>
#include <ec/esc_esi.h>
#include <ec/esc_mailbox.h>
#include <ec/esc_watchdog.h>
#include <ec/esc_registers.h>
#include <ec/esc_fmmu.h>
#include <log/log.h>

struct esc_device *esc_device_create2(struct ec_dgram_addr *ec_addr, char *name)
{
	struct esc_device *dev;
	
	log_send(LOG_T_DEBUG, "%s: Creating device", name);
	
	dev = calloc(1, sizeof(struct esc_device));
	if (dev)
		dev->name = strdup(name);

	memcpy(&dev->addr, ec_addr, sizeof(struct ec_dgram_addr));

	dev->mbox_rd_sm = -1;
	dev->mbox_wr_sm = -1;

	esc_init(dev);

	return dev;
}

struct esc_device *esc_device_create(char *busname, int devno, char *name)
{
	struct controller_bus *bus;
	struct ec_dgram_addr addr;
	struct esc_device *dev;
	
	ec_addr_set_auto_inc_nr(&addr, devno);

	bus = controller_bus_find(busname);
	if (strcmp(bus->name, "ethercat")) {
		log_send(LOG_T_ERROR, 
		    "Bus '%s' is not and ethercat bus, but of type '%s'",
		    bus->name, bus->type);
		return NULL;
	}
	
	dev = esc_device_create2(&addr, name);
	if (!dev)
		goto err_dev;

	dev->bus = bus;

	return dev;

err_dev:
	return NULL;
}

void esc_device_destroy(struct esc_device *dev)
{
	if (dev->categories)
		esc_esi_categories_destroy(dev->categories);
	if (dev->name)
		free(dev->name);
	
	free(dev);
}

struct controller_bus_client_private {
	struct esc_device dev;
};

static enum controller_bus_state esc_device_poll(struct controller_bus_client *client)
{
	struct esc_device *dev = &client->private->dev;
	enum controller_bus_state poll_state = CONTROLLER_BUS_STATE_OK;
	char *str;
	enum esc_al_status status;
	enum esc_al_state state;

	status = esc_al_status_code_get(dev);
	state = esc_al_state_get(dev);

	if (state == ESC_AL_STATE_ERROR || status < 0) {
		poll_state = CONTROLLER_BUS_STATE_ERROR;
	} else {
		if (state != dev->state) {
			log_send(LOG_T_WARNING,
			    "%s: ESC state changed: %d %s -> %d %s",
			    dev->name, 
			    dev->state, esc_al_state2str(dev->state),
			    state, esc_al_state2str(state));
			dev->state = state;
		}
		if (state != ESC_AL_STATE_OPERATIONAL) {
			poll_state = CONTROLLER_BUS_STATE_RECOVERABLE;
		}
		if (status != dev->status) {
			if (!status) {
				log_send(LOG_T_WARNING,
				    "%s: ESC status changed: No error", dev->name);
			} else {
				poll_state = CONTROLLER_BUS_STATE_RECOVERABLE;
				switch(status) {
					case ESC_AL_STATUS_SM_WATCHDOG:
						str = "Sync manager watchdog";
						break;
					case ESC_AL_STATUS_NO_ERROR:
					case ESC_AL_STATUS_UNSPECIFIED:
					case ESC_AL_STATUS_INVALID_REQ:
					case ESC_AL_STATUS_UNKNOWN_REQ:
					case ESC_AL_STATUS_BOOTSTRAP_UNSUP:
					case ESC_AL_STATUS_NO_VALID_FMW:
					case ESC_AL_STATUS_INVALID_MBOX1:
					case ESC_AL_STATUS_INVALID_MBOX2:
					case ESC_AL_STATUS_INVALID_SM:
					case ESC_AL_STATUS_NO_VALID_INPUT:
					case ESC_AL_STATUS_NO_VALID_OUTPUT:
					case ESC_AL_STATUS_SYNC_ERROR:
					case ESC_AL_STATUS_INVALID_SM_TYPES:
					case ESC_AL_STATUS_INVALID_OUT_CONF:
					case ESC_AL_STATUS_INVALID_IN_CONF:
					case ESC_AL_STATUS_INVALID_WD_CONF:
					case ESC_AL_STATUS_SLAVE_COLD:
					case ESC_AL_STATUS_SLAVE_INIT:
					case ESC_AL_STATUS_SLAVE_PREOP:
					case ESC_AL_STATUS_SLAVE_SAFEOP:
					case ESC_AL_STATUS_INVALID_DC_S_CONF:
					case ESC_AL_STATUS_INVALID_DC_L_CONF:
					case ESC_AL_STATUS_PLL_ERROR:
					case ESC_AL_STATUS_INVALID_DC_IO:
					case ESC_AL_STATUS_INVALID_DC_TO:
					case ESC_AL_STATUS_MBOX_EOE:
					case ESC_AL_STATUS_MBOX_COE:
					case ESC_AL_STATUS_MBOX_FOE:
					case ESC_AL_STATUS_MBOX_SOE:
					case ESC_AL_STATUS_MBOX_VOE:
					default:
						break;
				}
				log_send(LOG_T_ERROR, "%s: ESC status code changed: 0x%04x: %s",
				    dev->name, status, str);
			}
			dev->status = status;
		}
	}

	esc_al_error_ack(dev);
	return poll_state;
}

static int esc_device_recover(struct controller_bus_client *client)
{
	struct esc_device *dev = &client->private->dev;
	enum esc_al_state state;
	struct timespec	to = { 1, 0 };
	int r;

	state = esc_al_state_get(dev);
	log_send(LOG_T_WARNING, "cur state: %d", state);

	if (state == ESC_AL_STATE_ERROR) {
		return -1;
	}

	if (state == ESC_AL_STATE_INIT) {
		log_send(LOG_T_WARNING,
		    "%s: Trying to re-initialize to state %d %s",
		    dev->name, ESC_AL_STATE_PRE_OPERATIONAL,
		    esc_al_state2str(ESC_AL_STATE_PRE_OPERATIONAL));

		r = esc_device_initialize_pre_operational(dev);

		if (r) {
			log_send(LOG_T_ERROR,
			    "%s: Failed to re-initialize device",
			    dev->name);
			return r;
		} else {
			state = ESC_AL_STATE_PRE_OPERATIONAL;
		}
	}
	if (state == ESC_AL_STATE_PRE_OPERATIONAL) {
		log_send(LOG_T_WARNING, 
		    "%s: Trying to recover to state %d %s",
		    dev->name, ESC_AL_STATE_SAFE_OPERATIONAL,
		    esc_al_state2str(ESC_AL_STATE_SAFE_OPERATIONAL));

		if (dev->pre_operational_hook) {
			r = dev->pre_operational_hook(dev);
			if (r) {
				log_send(LOG_T_ERROR,
				   "%s: pre_operational_hook failed",
				   dev->name);
				return r;
			}
		}
			
		esc_al_error_ack(dev);
		state = esc_al_state_set(dev, ESC_AL_STATE_SAFE_OPERATIONAL, &to);
		if (state != ESC_AL_STATE_SAFE_OPERATIONAL) {
			return -1;
		}
	}
	if (state == ESC_AL_STATE_SAFE_OPERATIONAL) {
		log_send(LOG_T_WARNING, 
		    "%s: Trying to recover to state %d %s",
		    dev->name, ESC_AL_STATE_OPERATIONAL,
		    esc_al_state2str(ESC_AL_STATE_OPERATIONAL));

		esc_al_error_ack(dev);
		state = esc_al_state_set(dev, ESC_AL_STATE_OPERATIONAL, &to);
		if (state != ESC_AL_STATE_OPERATIONAL) {
			return -1;
		}
	}
	state = esc_al_state_get(dev);
	log_send(LOG_T_WARNING, "new state: %d", state);

	return 0;
}

/* initialize up to pre operational */
int esc_device_initialize_pre_operational(struct esc_device *dev)
{
	int r = 0;
	struct timespec	timeout = { 1, 0 };
	
	if (esc_al_state_set(dev, ESC_AL_STATE_INIT, &timeout) < 0)
		log_send(LOG_T_ERROR, "%s: Could not go to state init",
		    dev->name);

	if (dev->tx_pdo) {
		dev->sm[dev->tx_pdo_sm].ctrl =
		    ESC_SYNCMANAGER_CONTROL_WRITE |
		    ESC_SYNCMANAGER_CONTROL_INT_PDI |
		    (dev->tx_pdo_watchdog ? ESC_SYNCMANAGER_CONTROL_WATCHDOG_ENABLE : 0);

		esc_sm_set(dev, dev->tx_pdo_sm);

		if (dev->block_tx) {
			dev->client_tx = controller_bus_add_output_client(
			    dev->bus, dev->block_tx);
			if (dev->client_tx) {
				dev->client_tx->callback = dev->callback_tx;
				dev->client_tx->poll = esc_device_poll;
				dev->client_tx->recover = esc_device_recover;
				dev->client_tx->private = (struct controller_bus_client_private*)dev;
			}
			esc_fmmu_map_tx(dev, 
			    dev->sm[dev->tx_pdo_sm].start,
			    dev->sm[dev->tx_pdo_sm].len);
		}
	}
	if (dev->rx_pdo) {
		dev->sm[dev->rx_pdo_sm].ctrl =
		    ESC_SYNCMANAGER_CONTROL_INT_PDI;

		esc_sm_set(dev, dev->rx_pdo_sm);

		if (dev->block_rx) {
			dev->client_rx = controller_bus_add_input_client(
			    dev->bus, dev->block_rx);
			if (dev->client_rx) {
				dev->client_rx->callback = dev->callback_rx;
				dev->client_rx->poll = esc_device_poll;
				dev->client_rx->recover = esc_device_recover;
				dev->client_rx->private = (struct controller_bus_client_private*)dev;
			}
			esc_fmmu_map_rx(dev, 
			    dev->sm[dev->rx_pdo_sm].start,
			    dev->sm[dev->rx_pdo_sm].len);
		}
	}
	
	esc_fmmu_set(dev);
	

	if (dev->mbox_wr_sm >= 0 && dev->mbox_rd_sm >= 0) {
		dev->mailbox = esc_mailbox_create(&dev->addr, 
		    &dev->sm[dev->mbox_rd_sm], &dev->sm[dev->mbox_wr_sm]);
	}

	if (esc_al_state_set(dev, ESC_AL_STATE_PRE_OPERATIONAL, &timeout)< 0) {
		log_send(LOG_T_ERROR, 
		    "%s: Could not go to state pre operational", dev->name);
		r = -1;
	}

	esc_dc_init(&dev->addr, &dev->dc_sync);
	esc_watchdog_init(dev);

	dev->state = ESC_AL_STATE_PRE_OPERATIONAL;

	return r;
}

int esc_device_initialize_operational(struct esc_device *dev)
{
	int r;
	struct timespec	timeout = { 1, 0 };

	r = esc_device_initialize_pre_operational(dev);
	if (r)
		goto err;
	
	if (dev->pre_operational_hook) {
		r = dev->pre_operational_hook(dev);
		if (r)
			goto err;
	}
	
	r = esc_al_state_set(dev, ESC_AL_STATE_SAFE_OPERATIONAL, &timeout);
	if (r < 0) {
		log_send(LOG_T_ERROR, 
		    "%s: Could not go to state safe operational", dev->name);
		goto err;
	}
	dev->state = ESC_AL_STATE_SAFE_OPERATIONAL;
	
	r = esc_al_state_set(dev, ESC_AL_STATE_OPERATIONAL, &timeout);
	if (r < 0) {
		log_send(LOG_T_ERROR, 
		    "%s: Could not go to state operational", dev->name);
		goto err;
	}
	dev->state = ESC_AL_STATE_OPERATIONAL;
	
err:
	return r;
}
